@@ -6,6 +6,7 @@ import (
 	"github.com/containernetworking/cni/pkg/skel"
 	"github.com/containernetworking/cni/pkg/types"
 	"github.com/containernetworking/cni/pkg/version"
+	bv "github.com/containernetworking/plugins/pkg/utils/buildversion"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/cni"
 	"github.com/urfave/cli"
 )
@@ -20,8 +21,10 @@ func main() {
 	c.Action = func(ctx *cli.Context) error {
 		skel.PluginMain(
 			p.CmdAdd,
+			p.CmdCheck,
 			p.CmdDel,
-			version.All)
+			version.PluginSupports("0.1.0", "0.2.0", "0.3.0", "0.3.1"),
+			bv.BuildString("ovn-k8s-cni-overlay"))
 		return nil
 	}
 
