@@ -4,31 +4,20 @@ import (
 	"fmt"
 	"net"
 
-	"github.com/openshift/origin/pkg/util/netutils"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/factory"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/kube"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
+	kapi "k8s.io/api/core/v1"
 	"k8s.io/client-go/kubernetes"
 )
 
 // OvnClusterController is the object holder for utilities meant for cluster management
 type OvnClusterController struct {
-	Kube                      kube.Interface
-	watchFactory              *factory.WatchFactory
-	masterSubnetAllocatorList []*netutils.SubnetAllocator
-
-	TCPLoadBalancerUUID string
-	UDPLoadBalancerUUID string
+	Kube         kube.Interface
+	watchFactory *factory.WatchFactory
 }
 
-const (
-	// OvnHostSubnet is the constant string representing the annotation key
-	OvnHostSubnet = "ovn_host_subnet"
-	// OvnClusterRouter is the name of the distributed router
-	OvnClusterRouter = "ovn_cluster_router"
-)
-
 // NewClusterController creates a new controller for IP subnet allocation to
 // a given resource type (either Namespace or Node)
 func NewClusterController(kubeClient kubernetes.Interface, wf *factory.WatchFactory) *OvnClusterController {
@@ -38,21 +27,19 @@ func NewClusterController(kubeClient kubernetes.Interface, wf *factory.WatchFact
 	}
 }
 
-func setupOVNNode(nodeName string) error {
-	// Tell ovn-*bctl how to talk to the database
-	for _, auth := range []config.OvnAuthConfig{config.OvnNorth, config.OvnSouth} {
-		if err := auth.SetDBAuth(); err != nil {
-			return err
-		}
-	}
-
+func setupOVNNode(node *kapi.Node) error {
 	var err error
 
+	nodeName, err := util.GetNodeHostname(node)
+	if err != nil {
+		return fmt.Errorf("failed to obtain hostname from node %q: %v", node.Name, err)
+	}
+
 	nodeIP := config.Default.EncapIP
 	if nodeIP == "" {
-		nodeIP, err = netutils.GetNodeIP(nodeName)
+		nodeIP, err = util.GetNodeIP(node)
 		if err != nil {
-			return fmt.Errorf("failed to obtain local IP from hostname %q: %v", nodeName, err)
+			return fmt.Errorf("failed to obtain local IP from node %q: %v", node.Name, err)
 		}
 	} else {
 		if ip := net.ParseIP(nodeIP); ip == nil {
@@ -67,20 +54,33 @@ func setupOVNNode(nodeName string) error {
 		fmt.Sprintf("external_ids:ovn-encap-ip=%s", nodeIP),
 		fmt.Sprintf("external_ids:ovn-remote-probe-interval=%d",
 			config.Default.InactivityProbe),
+		fmt.Sprintf("external_ids:ovn-openflow-probe-interval=%d",
+			config.Default.OpenFlowProbe),
 		fmt.Sprintf("external_ids:hostname=\"%s\"", nodeName),
 	)
 	if err != nil {
 		return fmt.Errorf("error setting OVS external IDs: %v\n  %q", err, stderr)
 	}
-	return nil
-}
-
-func setupOVNMaster(nodeName string) error {
-	// Configure both server and client of OVN databases, since master uses both
-	for _, auth := range []config.OvnAuthConfig{config.OvnNorth, config.OvnSouth} {
-		if err := auth.SetDBAuth(); err != nil {
+	// If EncapPort is not the default tell sbdb to use specified port.
+	if config.Default.EncapPort != config.DefaultEncapPort {
+		systemID, err := util.GetNodeChassisID()
+		if err != nil {
 			return err
 		}
+		uuid, _, err := util.RunOVNSbctl("--data=bare", "--no-heading", "--columns=_uuid", "find", "Encap",
+			fmt.Sprintf("chassis_name=%s", systemID))
+		if err != nil {
+			return err
+		}
+		if len(uuid) == 0 {
+			return fmt.Errorf("unable to find encap uuid to set geneve port for chassis %s", systemID)
+		}
+		_, stderr, errSet := util.RunOVNSbctl("set", "encap", uuid,
+			fmt.Sprintf("options:dst_port=%d", config.Default.EncapPort),
+		)
+		if errSet != nil {
+			return fmt.Errorf("error setting OVS encap-port: %v\n  %q", errSet, stderr)
+		}
 	}
 	return nil
 }
