@@ -7,6 +7,7 @@ import (
 
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/factory"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/ovn"
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
 
 	"github.com/sirupsen/logrus"
@@ -61,16 +62,7 @@ func deleteService(service *kapi.Service, inport, gwBridge string) {
 	}
 }
 
-func syncServices(services []interface{}, gwBridge, gwIntf string) {
-	// Get ofport of physical interface
-	inport, stderr, err := util.RunOVSVsctl("--if-exists", "get",
-		"interface", gwIntf, "ofport")
-	if err != nil {
-		logrus.Errorf("Failed to get ofport of %s, stderr: %q, error: %v",
-			gwIntf, stderr, err)
-		return
-	}
-
+func syncServices(services []interface{}, inport, gwBridge string) {
 	nodePorts := make(map[string]bool)
 	for _, serviceInterface := range services {
 		service, ok := serviceInterface.(*kapi.Service)
@@ -178,7 +170,7 @@ func nodePortWatcher(nodeName, gwBridge, gwIntf string, wf *factory.WatchFactory
 			deleteService(service, ofportPhys, gwBridge)
 		},
 	}, func(services []interface{}) {
-		syncServices(services, gwBridge, gwIntf)
+		syncServices(services, ofportPhys, gwBridge)
 	})
 
 	return err
@@ -243,9 +235,9 @@ func addDefaultConntrackRules(nodeName, gwBridge, gwIntf string) error {
 			"error: %v", gwBridge, stderr, err)
 	}
 
-	// table 1, all other connections go to the bridge interface.
+	// table 1, all other connections do normal processing
 	_, stderr, err = util.RunOVSOfctl("add-flow", gwBridge,
-		"priority=0, table=1, actions=output:LOCAL")
+		"priority=0, table=1, actions=output:NORMAL")
 	if err != nil {
 		return fmt.Errorf("Failed to add openflow flow to %s, stderr: %q, "+
 			"error: %v", gwBridge, stderr, err)
@@ -253,100 +245,83 @@ func addDefaultConntrackRules(nodeName, gwBridge, gwIntf string) error {
 	return nil
 }
 
-func addStaticRouteToHost(nodeName, nicIP string) error {
-	k8sClusterRouter, err := util.GetK8sClusterRouter()
-	if err != nil {
-		return err
-	}
-
-	k8sMgmtIntfName := util.GetK8sMgmtIntfName(nodeName)
-	k8sMgmtIntfIPAddress, err := getIPv4Address(k8sMgmtIntfName)
-	if err != nil {
-		return fmt.Errorf("failed to get interface IP address for %s (%v)",
-			k8sMgmtIntfName, err)
-	}
-	prefix := strings.Split(nicIP, "/")[0] + "/32"
-	nexthop := strings.Split(k8sMgmtIntfIPAddress, "/")[0]
-	_, stderr, err := util.RunOVNNbctl("--may-exist", "lr-route-add", k8sClusterRouter, prefix, nexthop)
-	if err != nil {
-		return fmt.Errorf("failed to add static route '%s via %s' for host %q on %s "+
-			"stderr: %q, error: %v", nicIP, k8sMgmtIntfIPAddress, nodeName, k8sClusterRouter, stderr, err)
-	}
-	return nil
-}
-
-func initSharedGateway(
-	nodeName string, clusterIPSubnet []string, subnet,
-	gwNextHop, gwIntf string, wf *factory.WatchFactory) error {
+func initSharedGateway(nodeName string, subnet, gwNextHop, gwIntf string,
+	wf *factory.WatchFactory) (map[string]map[string]string, postReadyFn, error) {
 	var bridgeName string
+	var uplinkName string
+	var brCreated bool
+	var err error
 
-	// Check to see whether the interface is OVS bridge.
-	if _, _, err := util.RunOVSVsctl("--", "br-exists", gwIntf); err != nil {
+	if bridgeName, _, err = util.RunOVSVsctl("--", "port-to-br", gwIntf); err == nil {
+		// This is an OVS bridge's internal port
+		uplinkName, err = util.GetNicName(bridgeName)
+		if err != nil {
+			return nil, nil, err
+		}
+	} else if _, _, err := util.RunOVSVsctl("--", "br-exists", gwIntf); err != nil {
 		// This is not a OVS bridge. We need to create a OVS bridge
 		// and add cluster.GatewayIntf as a port of that bridge.
 		bridgeName, err = util.NicToBridge(gwIntf)
 		if err != nil {
-			return fmt.Errorf("failed to convert %s to OVS bridge: %v",
+			return nil, nil, fmt.Errorf("failed to convert %s to OVS bridge: %v",
 				gwIntf, err)
 		}
+		uplinkName = gwIntf
+		gwIntf = bridgeName
+		brCreated = true
 	} else {
-		intfName, err := getIntfName(gwIntf)
+		// gateway interface is an OVS bridge
+		uplinkName, err = getIntfName(gwIntf)
 		if err != nil {
-			return err
+			return nil, nil, err
 		}
 		bridgeName = gwIntf
-		gwIntf = intfName
 	}
 
 	// Now, we get IP address from OVS bridge. If IP does not exist,
 	// error out.
-	ipAddress, err := getIPv4Address(bridgeName)
+	ipAddress, err := getIPv4Address(gwIntf)
 	if err != nil {
-		return fmt.Errorf("Failed to get interface details for %s (%v)",
-			bridgeName, err)
+		return nil, nil, fmt.Errorf("Failed to get interface details for %s (%v)",
+			gwIntf, err)
 	}
 	if ipAddress == "" {
-		return fmt.Errorf("%s does not have a ipv4 address", bridgeName)
+		return nil, nil, fmt.Errorf("%s does not have a ipv4 address", gwIntf)
 	}
 
-	ifaceID, macAddress, err := bridgedGatewayNodeSetup(nodeName, bridgeName)
+	ifaceID, macAddress, err := bridgedGatewayNodeSetup(nodeName, bridgeName, gwIntf, brCreated)
 	if err != nil {
-		return fmt.Errorf("failed to set up shared interface gateway: %v", err)
+		return nil, nil, fmt.Errorf("failed to set up shared interface gateway: %v", err)
 	}
 
-	var lspArgs []string
-	if config.Gateway.VLANID > 0 {
-		lspArgs = []string{"--", "set", "logical_switch_port",
-			ifaceID, fmt.Sprintf("tag_request=%d", config.Gateway.VLANID)}
+	l3GatewayConfig := map[string]string{
+		ovn.OvnNodeGatewayMode:       string(config.Gateway.Mode),
+		ovn.OvnNodeGatewayVlanID:     fmt.Sprintf("%d", config.Gateway.VLANID),
+		ovn.OvnNodeGatewayIfaceID:    ifaceID,
+		ovn.OvnNodeGatewayMacAddress: macAddress,
+		ovn.OvnNodeGatewayIP:         ipAddress,
+		ovn.OvnNodeGatewayNextHop:    gwNextHop,
+		ovn.OvnNodePortEnable:        fmt.Sprintf("%t", config.Gateway.NodeportEnable),
+	}
+	annotations := map[string]map[string]string{
+		ovn.OvnDefaultNetworkGateway: l3GatewayConfig,
 	}
 
-	err = util.GatewayInit(clusterIPSubnet, nodeName, ifaceID, ipAddress,
-		macAddress, gwNextHop, subnet, true, lspArgs)
-	if err != nil {
-		return fmt.Errorf("failed to init shared interface gateway: %v", err)
-	}
-
-	// Add static routes to OVN Cluster Router to enable pods on this Node to
-	// reach the host IP
-	err = addStaticRouteToHost(nodeName, ipAddress)
-	if err != nil {
-		return err
-	}
-
-	// Program cluster.GatewayIntf to let non-pod traffic to go to host
-	// stack
-	if err := addDefaultConntrackRules(nodeName, bridgeName, gwIntf); err != nil {
-		return err
-	}
-
-	if config.Gateway.NodeportEnable {
-		// Program cluster.GatewayIntf to let nodePort traffic to go to pods.
-		if err := nodePortWatcher(nodeName, bridgeName, gwIntf, wf); err != nil {
+	return annotations, func() error {
+		// Program cluster.GatewayIntf to let non-pod traffic to go to host
+		// stack
+		if err := addDefaultConntrackRules(nodeName, bridgeName, uplinkName); err != nil {
 			return err
 		}
-	}
 
-	return nil
+		if config.Gateway.NodeportEnable {
+			// Program cluster.GatewayIntf to let nodePort traffic to go to pods.
+			if err := nodePortWatcher(nodeName, bridgeName, uplinkName, wf); err != nil {
+				return err
+			}
+		}
+		return nil
+	}, nil
 }
 
 func cleanupSharedGateway() error {
@@ -363,5 +338,30 @@ func cleanupSharedGateway() error {
 			return fmt.Errorf("Failed to delete port %s stderr:%s (%v)", port, stderr, err)
 		}
 	}
+
+	// Get the OVS bridge name from ovn-bridge-mappings
+	stdout, stderr, err = util.RunOVSVsctl("--if-exists", "get", "Open_vSwitch", ".",
+		"external_ids:ovn-bridge-mappings")
+	if err != nil {
+		return fmt.Errorf("Failed to get ovn-bridge-mappings stderr:%s (%v)", stderr, err)
+	}
+	// skip the existing mapping setting for the specified physicalNetworkName
+	bridgeName := ""
+	bridgeMappings := strings.Split(stdout, ",")
+	for _, bridgeMapping := range bridgeMappings {
+		m := strings.Split(bridgeMapping, ":")
+		if network := m[0]; network == util.PhysicalNetworkName {
+			bridgeName = m[1]
+			break
+		}
+	}
+	if len(bridgeName) == 0 {
+		return nil
+	}
+
+	_, stderr, err = util.AddNormalActionOFFlow(bridgeName)
+	if err != nil {
+		return fmt.Errorf("Failed to replace-flows on bridge %q stderr:%s (%v)", bridgeName, stderr, err)
+	}
 	return nil
 }
