@@ -5,7 +5,7 @@ package cluster
 import (
 	"fmt"
 
-	"github.com/openvswitch/ovn-kubernetes/go-controller/pkg/util"
+	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
 )
 
 // getDefaultGatewayInterfaceDetails returns the interface name on
