@@ -23,11 +23,14 @@ import (
 // attach it to the desired network. This function finds the HNS Id of the
 // network based on the gatewayIP. If more than one suitable network it's found,
 // return an error asking to give the HNS Network Id in config.
-func getHNSIdFromConfigOrByGatewayIP(gatewayIP string) (string, error) {
+func getHNSIdFromConfigOrByGatewayIP(gatewayIP net.IP) (string, error) {
 	if config.CNI.WinHNSNetworkID != "" {
 		logrus.Infof("Using HNS Network Id from config: %v", config.CNI.WinHNSNetworkID)
 		return config.CNI.WinHNSNetworkID, nil
 	}
+	if gatewayIP == nil {
+		return "", fmt.Errorf("no gateway IP and no HNS Network ID given")
+	}
 	hnsNetworkId := ""
 	hnsNetworks, err := hcsshim.HNSListNetworkRequest("GET", "", "")
 	if err != nil {
@@ -35,7 +38,7 @@ func getHNSIdFromConfigOrByGatewayIP(gatewayIP string) (string, error) {
 	}
 	for _, hnsNW := range hnsNetworks {
 		for _, hnsNWSubnet := range hnsNW.Subnets {
-			if strings.Compare(gatewayIP, hnsNWSubnet.GatewayAddress) == 0 {
+			if strings.Compare(gatewayIP.String(), hnsNWSubnet.GatewayAddress) == 0 {
 				if len(hnsNetworkId) == 0 {
 					hnsNetworkId = hnsNW.Id
 				} else {
@@ -109,23 +112,20 @@ func deleteHNSEndpoint(endpointName string) error {
 // The fact that CNI add should be idempotent on Windows is stated here:
 // https://github.com/kubernetes/kubernetes/blob/master/pkg/kubelet/network/cni/cni_windows.go#L38
 // TODO: add proper MTU config (GetCurrentThreadId/SetCurrentThreadId) or via OVS properties
-func (pr *PodRequest) ConfigureInterface(namespace string, podName string, macAddress string, ipAddress string, gatewayIP string, mtu int, ingress, egress int64) ([]*current.Interface, error) {
+func (pr *PodRequest) ConfigureInterface(namespace string, podName string, ifInfo *PodInterfaceInfo) ([]*current.Interface, error) {
 	conf := pr.CNIConf
 
 	if conf.DeviceID != "" {
 		return nil, fmt.Errorf("failure OVS-Offload is not supported in Windows")
 	}
-	ipAddr, ipNet, err := net.ParseCIDR(ipAddress)
-	if err != nil {
-		return nil, err
-	}
-	ipMaskSize, _ := ipNet.Mask.Size()
+	ipMaskSize, _ := ifInfo.IP.Mask.Size()
 	// NOTE(abalutoiu): The endpoint name should not depend on the container ID.
 	// This is for backwards compatibility in kubernetes which calls the CNI
 	// even for containers that are not the infra container.
 	// This is getting fixed by https://github.com/kubernetes/kubernetes/pull/64189
 	endpointName := fmt.Sprintf("%s_%s", namespace, podName)
 
+	var err error
 	defer func() {
 		// Delete the endpoint in case CNI fails
 		if err != nil {
@@ -137,7 +137,7 @@ func (pr *PodRequest) ConfigureInterface(namespace string, podName string, macAd
 	}()
 
 	var hnsNetworkId string
-	hnsNetworkId, err = getHNSIdFromConfigOrByGatewayIP(gatewayIP)
+	hnsNetworkId, err = getHNSIdFromConfigOrByGatewayIP(ifInfo.GW)
 	if err != nil {
 		logrus.Infof("Error when detecting the HNS Network Id: %q", err)
 		return nil, err
@@ -151,12 +151,12 @@ func (pr *PodRequest) ConfigureInterface(namespace string, podName string, macAd
 		logrus.Infof("HNS endpoint %q does not exist", endpointName)
 
 		// HNSEndpoint requires the xx-xx-xx-xx-xx-xx format for the MacAddress field
-		macAddressIpFormat := strings.Replace(macAddress, ":", "-", -1)
+		macAddressIpFormat := strings.Replace(ifInfo.MAC.String(), ":", "-", -1)
 
 		hnsEndpoint := &hcsshim.HNSEndpoint{
 			Name:           endpointName,
 			VirtualNetwork: hnsNetworkId,
-			IPAddress:      ipAddr,
+			IPAddress:      ifInfo.IP.IP,
 			MacAddress:     macAddressIpFormat,
 			PrefixLength:   uint8(ipMaskSize),
 			DNSServerList:  strings.Join(conf.DNS.Nameservers, ","),
@@ -194,9 +194,9 @@ func (pr *PodRequest) ConfigureInterface(namespace string, podName string, macAd
 		"--may-exist", "add-port", "br-int", endpointName, "--", "set",
 		"interface", endpointName, "type=internal", "--", "set",
 		"interface", endpointName,
-		fmt.Sprintf("external_ids:attached_mac=%s", macAddress),
+		fmt.Sprintf("external_ids:attached_mac=%s", ifInfo.MAC),
 		fmt.Sprintf("external_ids:iface-id=%s", ifaceID),
-		fmt.Sprintf("external_ids:ip_address=%s", ipAddress),
+		fmt.Sprintf("external_ids:ip_address=%s", ifInfo.IP),
 	}
 	var out []byte
 	out, err = exec.Command("ovs-vsctl", ovsArgs...).CombinedOutput()
@@ -208,7 +208,7 @@ func (pr *PodRequest) ConfigureInterface(namespace string, podName string, macAd
 		"Set-NetIPInterface",
 		"-IncludeAllCompartments",
 		fmt.Sprintf("-InterfaceAlias \"vEthernet (%s)\"", ifaceID),
-		fmt.Sprintf("-NlMtuBytes %d", mtu),
+		fmt.Sprintf("-NlMtuBytes %d", ifInfo.MTU),
 	}
 
 	out, err = exec.Command("powershell", mtuArgs...).CombinedOutput()
