@@ -11,13 +11,22 @@ import (
 	"runtime"
 	"strings"
 
+	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+
 	"github.com/sirupsen/logrus"
 	"github.com/urfave/cli"
 	gcfg "gopkg.in/gcfg.v1"
+	lumberjack "gopkg.in/natefinch/lumberjack.v2"
 
 	kexec "k8s.io/utils/exec"
 )
 
+// DefaultEncapPort number used if not supplied
+const DefaultEncapPort = 6081
+
+const MetricNamespace = "ovnkube"
+const DefaultAPIServer = "http://localhost:8443"
+
 // The following are global config parameters that other modules may access directly
 var (
 	// ovn-kubernetes version, to be changed with every release
@@ -32,7 +41,9 @@ var (
 		ConntrackZone:     64000,
 		EncapType:         "geneve",
 		EncapIP:           "",
-		InactivityProbe:   100000,
+		EncapPort:         DefaultEncapPort,
+		InactivityProbe:   100000, // in Milliseconds
+		OpenFlowProbe:     180,    // in Seconds
 		RawClusterSubnets: "10.128.0.0/14/23",
 	}
 
@@ -51,7 +62,7 @@ var (
 
 	// Kubernetes holds Kubernetes-related parsed config file parameters and command-line overrides
 	Kubernetes = KubernetesConfig{
-		APIServer:          "http://localhost:8080",
+		APIServer:          DefaultAPIServer,
 		ServiceCIDR:        "172.16.1.0/24",
 		OVNConfigNamespace: "ovn-kubernetes",
 	}
@@ -64,6 +75,28 @@ var (
 
 	// Gateway holds node gateway-related parsed config file parameters and command-line overrides
 	Gateway GatewayConfig
+
+	// MasterHA holds master HA related config options.
+	MasterHA = MasterHAConfig{
+		ElectionLeaseDuration: 60,
+		ElectionRenewDeadline: 30,
+		ElectionRetryPeriod:   20,
+		ManageDBServers:       false,
+		NbPort:                6641,
+		SbPort:                6642,
+	}
+
+	// NbctlDaemon enables ovn-nbctl to run in daemon mode
+	NbctlDaemonMode bool
+
+	// UnprivilegedMode allows ovnkube-node to run without SYS_ADMIN capability, by performing interface setup in the CNI plugin
+	UnprivilegedMode bool
+
+	// EnableMulticast enables multicast support between the pods within the same namespace
+	EnableMulticast bool
+
+	// IPv6Mode captures whether we are using IPv6 for OVN logical topology
+	IPv6Mode bool
 )
 
 const (
@@ -86,9 +119,15 @@ type DefaultConfig struct {
 	// The IP address of the encapsulation endpoint. If not specified, the IP address the
 	// NodeName resolves to will be used
 	EncapIP string `gcfg:"encap-ip"`
+	// The UDP Port of the encapsulation endpoint. If not specified, the IP default port
+	// of 6081 will be used
+	EncapPort uint `gcfg:"encap-port"`
 	// Maximum number of milliseconds of idle time on connection that
 	// ovn-controller waits before it will send a connection health probe.
 	InactivityProbe int `gcfg:"inactivity-probe"`
+	// Maximum number of seconds of idle time on the OpenFlow connection
+	// that ovn-controller will wait before it sends a connection health probe
+	OpenFlowProbe int `gcfg:"openflow-probe"`
 	// RawClusterSubnets holds the unparsed cluster subnets. Should only be
 	// used inside config module.
 	RawClusterSubnets string `gcfg:"cluster-subnets"`
@@ -117,12 +156,18 @@ type CNIConfig struct {
 
 // KubernetesConfig holds Kubernetes-related parsed config file parameters and command-line overrides
 type KubernetesConfig struct {
-	Kubeconfig         string `gcfg:"kubeconfig"`
-	CACert             string `gcfg:"cacert"`
-	APIServer          string `gcfg:"apiserver"`
-	Token              string `gcfg:"token"`
-	ServiceCIDR        string `gcfg:"service-cidr"`
-	OVNConfigNamespace string `gcfg:"ovn-config-namespace"`
+	Kubeconfig           string `gcfg:"kubeconfig"`
+	CACert               string `gcfg:"cacert"`
+	APIServer            string `gcfg:"apiserver"`
+	Token                string `gcfg:"token"`
+	ServiceCIDR          string `gcfg:"service-cidr"`
+	OVNConfigNamespace   string `gcfg:"ovn-config-namespace"`
+	MetricsBindAddress   string `gcfg:"metrics-bind-address"`
+	MetricsEnablePprof   bool   `gcfg:"metrics-enable-pprof"`
+	OVNEmptyLbEvents     bool   `gcfg:"ovn-empty-lb-events"`
+	PodIP                string `gcfg:"pod-ip"`
+	RawNoHostSubnetNodes string `gcfg:"no-hostsubnet-nodes"`
+	NoHostSubnetNodes    *metav1.LabelSelector
 }
 
 // GatewayMode holds the node gateway mode
@@ -133,21 +178,19 @@ const (
 	GatewayModeDisabled GatewayMode = ""
 	// GatewayModeShared indicates OVN shares a gateway interface with the node
 	GatewayModeShared GatewayMode = "shared"
-	// GatewayModeSpare indicates OVN claims a spare interfaface for the gateway
-	GatewayModeSpare GatewayMode = "spare"
 	// GatewayModeLocal indicates OVN creates a local NAT-ed interface for the gateway
 	GatewayModeLocal GatewayMode = "local"
 )
 
 // GatewayConfig holds node gateway-related parsed config file parameters and command-line overrides
 type GatewayConfig struct {
-	// Mode is the gateway mode; if may be either empty (disabled), "shared", "spare", or "local"
+	// Mode is the gateway mode; if may be either empty (disabled), "shared", or "local"
 	Mode GatewayMode `gcfg:"mode"`
-	// Interface is the network interface to use for the gateway in "shared" or "spare" mode
+	// Interface is the network interface to use for the gateway in "shared" mode
 	Interface string `gcfg:"interface"`
 	// NextHop is the gateway IP address of Interface; will be autodetected if not given
 	NextHop string `gcfg:"next-hop"`
-	// VLANID is the option VLAN tag to apply to gateway traffic for "shared" or "spare" modes
+	// VLANID is the option VLAN tag to apply to gateway traffic for "shared" mode
 	VLANID uint `gcfg:"vlan-id"`
 	// NodeportEnable sets whether to provide Kubernetes NodePort service or not
 	NodeportEnable bool `gcfg:"nodeport"`
@@ -169,6 +212,17 @@ type OvnAuthConfig struct {
 	exec kexec.Interface
 }
 
+// MasterHAConfig holds configuration for master HA
+// configuration.
+type MasterHAConfig struct {
+	ElectionLeaseDuration int  `gcfg:"election-lease-duration"`
+	ElectionRenewDeadline int  `gcfg:"election-renew-deadline"`
+	ElectionRetryPeriod   int  `gcfg:"election-retry-period"`
+	ManageDBServers       bool `gcfg:"manage-db-servers"`
+	NbPort                int  `gcfg:"port"`
+	SbPort                int  `gcfg:"port"`
+}
+
 // OvnDBScheme describes the OVN database connection transport method
 type OvnDBScheme string
 
@@ -190,6 +244,7 @@ type config struct {
 	OvnNorth   OvnAuthConfig
 	OvnSouth   OvnAuthConfig
 	Gateway    GatewayConfig
+	MasterHA   MasterHAConfig
 }
 
 var (
@@ -200,21 +255,19 @@ var (
 	savedOvnNorth   OvnAuthConfig
 	savedOvnSouth   OvnAuthConfig
 	savedGateway    GatewayConfig
-
+	savedMasterHA   MasterHAConfig
 	// legacy service-cluster-ip-range CLI option
 	serviceClusterIPRange string
 	// legacy cluster-subnet CLI option
 	clusterSubnet string
 	// legacy init-gateways CLI option
 	initGateways bool
-	// legacy gateway-spare-interface CLI option
-	gatewaySpareInterface bool
 	// legacy gateway-local CLI option
 	gatewayLocal bool
 )
 
 func init() {
-	// Cache original default config values so they can be restored by testcases
+	// Cache original default config values
 	savedDefault = Default
 	savedLogging = Logging
 	savedCNI = CNI
@@ -222,12 +275,14 @@ func init() {
 	savedOvnNorth = OvnNorth
 	savedOvnSouth = OvnSouth
 	savedGateway = Gateway
+	savedMasterHA = MasterHA
 	Flags = append(Flags, CommonFlags...)
 	Flags = append(Flags, CNIFlags...)
 	Flags = append(Flags, K8sFlags...)
 	Flags = append(Flags, OvnNBFlags...)
 	Flags = append(Flags, OvnSBFlags...)
 	Flags = append(Flags, OVNGatewayFlags...)
+	Flags = append(Flags, MasterHAFlags...)
 }
 
 // RestoreDefaultConfig restores default config values. Used by testcases to
@@ -240,21 +295,26 @@ func RestoreDefaultConfig() {
 	OvnNorth = savedOvnNorth
 	OvnSouth = savedOvnSouth
 	Gateway = savedGateway
+	MasterHA = savedMasterHA
 }
 
 // copy members of struct 'src' into the corresponding field in struct 'dst'
-// if the field in 'src' is a non-zero int or a non-zero-length string. This
-// function should be called with pointers to structs.
-func overrideFields(dst, src interface{}) {
+// if the field in 'src' is a non-zero int or a non-zero-length string and
+// does not contain a default value. This function should be called with pointers to structs.
+func overrideFields(dst, src, defaults interface{}) error {
 	dstStruct := reflect.ValueOf(dst).Elem()
 	srcStruct := reflect.ValueOf(src).Elem()
 	if dstStruct.Kind() != srcStruct.Kind() || dstStruct.Kind() != reflect.Struct {
-		panic("mismatched value types")
+		return fmt.Errorf("mismatched value types")
 	}
 	if dstStruct.NumField() != srcStruct.NumField() {
-		panic("mismatched struct types")
+		return fmt.Errorf("mismatched struct types")
 	}
 
+	var defStruct reflect.Value
+	if defaults != nil {
+		defStruct = reflect.ValueOf(defaults).Elem()
+	}
 	// Iterate over each field in dst/src Type so we can get the tags,
 	// and use the field name to retrieve the field's actual value from
 	// the dst/src instance
@@ -271,40 +331,27 @@ func overrideFields(dst, src interface{}) {
 
 		dstField := dstStruct.FieldByName(structField.Name)
 		srcField := srcStruct.FieldByName(structField.Name)
+		var dv reflect.Value
+		if defStruct.IsValid() {
+			dv = defStruct.FieldByName(structField.Name)
+		}
 		if !dstField.IsValid() || !srcField.IsValid() {
-			panic(fmt.Sprintf("invalid struct %q field %q",
-				dstType.Name(), structField.Name))
+			return fmt.Errorf("invalid struct %q field %q", dstType.Name(), structField.Name)
 		}
 		if dstField.Kind() != srcField.Kind() {
-			panic(fmt.Sprintf("mismatched struct %q fields %q",
-				dstType.Name(), structField.Name))
+			return fmt.Errorf("mismatched struct %q fields %q", dstType.Name(), structField.Name)
 		}
-		switch srcField.Kind() {
-		case reflect.String:
-			if srcField.String() != "" {
-				dstField.Set(srcField)
-			}
-		case reflect.Int:
-			if srcField.Int() != 0 {
-				dstField.Set(srcField)
-			}
-		case reflect.Uint:
-			if srcField.Uint() != 0 {
-				dstField.Set(srcField)
-			}
-		case reflect.Bool:
-			if srcField.Bool() {
-				dstField.Set(srcField)
-			}
-		default:
-			panic(fmt.Sprintf("unhandled struct %q field %q type %v",
-				dstType.Name(), structField.Name, srcField.Kind()))
+		if dv.IsValid() && reflect.DeepEqual(dv.Interface(), srcField.Interface()) {
+			continue
 		}
+		dstField.Set(srcField)
 	}
 	if !handled {
 		// No tags found in the struct so we don't know how to override
-		panic(fmt.Sprintf("failed to find 'gcfg' tags in struct %q", dstType.Name()))
+		return fmt.Errorf("failed to find 'gcfg' tags in struct %q", dstType.Name())
 	}
+
+	return nil
 }
 
 var cliConfig config
@@ -331,32 +378,50 @@ var CommonFlags = []cli.Flag{
 	cli.StringFlag{
 		Name:  "config-file",
 		Usage: "configuration file path (default: /etc/openvswitch/ovn_k8s.conf)",
+		//Value: "/etc/openvswitch/ovn_k8s.conf",
 	},
 	cli.IntFlag{
 		Name:        "mtu",
 		Usage:       "MTU value used for the overlay networks (default: 1400)",
 		Destination: &cliConfig.Default.MTU,
+		Value:       Default.MTU,
 	},
 	cli.IntFlag{
 		Name:        "conntrack-zone",
 		Usage:       "For gateway nodes, the conntrack zone used for conntrack flow rules (default: 64000)",
 		Destination: &cliConfig.Default.ConntrackZone,
+		Value:       Default.ConntrackZone,
 	},
 	cli.StringFlag{
 		Name:        "encap-type",
 		Usage:       "The encapsulation protocol to use to transmit packets between hypervisors (default: geneve)",
 		Destination: &cliConfig.Default.EncapType,
+		Value:       Default.EncapType,
 	},
 	cli.StringFlag{
 		Name:        "encap-ip",
 		Usage:       "The IP address of the encapsulation endpoint (default: Node IP address resolved from Node hostname)",
 		Destination: &cliConfig.Default.EncapIP,
 	},
+	cli.UintFlag{
+		Name:        "encap-port",
+		Usage:       "The UDP port used by the encapsulation endpoint (default: 6081)",
+		Destination: &cliConfig.Default.EncapPort,
+		Value:       Default.EncapPort,
+	},
 	cli.IntFlag{
 		Name: "inactivity-probe",
 		Usage: "Maximum number of milliseconds of idle time on " +
 			"connection for ovn-controller before it sends a inactivity probe",
 		Destination: &cliConfig.Default.InactivityProbe,
+		Value:       Default.InactivityProbe,
+	},
+	cli.IntFlag{
+		Name: "openflow-probe",
+		Usage: "Maximum number of seconds of idle time on the openflow " +
+			"connection for ovn-controller before it sends a inactivity probe",
+		Destination: &cliConfig.Default.OpenFlowProbe,
+		Value:       Default.OpenFlowProbe,
 	},
 	cli.StringFlag{
 		Name:        "cluster-subnet",
@@ -365,23 +430,39 @@ var CommonFlags = []cli.Flag{
 	},
 	cli.StringFlag{
 		Name:  "cluster-subnets",
-		Value: "10.128.0.0/14/23",
+		Value: Default.RawClusterSubnets,
 		Usage: "A comma separated set of IP subnets and the associated " +
 			"hostsubnet prefix lengths to use for the cluster (eg, \"10.128.0.0/14/23,10.0.0.0/14/23\"). " +
 			"Each entry is given in the form [IP address/prefix-length/hostsubnet-prefix-length] " +
 			"and cannot overlap with other entries. The hostsubnet-prefix-length " +
-			"is optional and if unspecified defaults to 24. The " +
-			"hostsubnet-prefix-length defines how many IP addresses " +
-			"are dedicated to each node and may be different for each " +
-			"entry.",
+			"defines how large a subnet is given to each node and may be different " +
+			"for each entry. For IPv6 subnets, it must be 64 (and does not need to " +
+			"be explicitly specified). For IPv4 subnets an explicit " +
+			"hostsubnet-prefix should be specified, but for backward compatibility " +
+			"it defaults to 24 if unspecified.",
 		Destination: &cliConfig.Default.RawClusterSubnets,
 	},
-
+	cli.BoolFlag{
+		Name:        "nbctl-daemon-mode",
+		Usage:       "Run ovn-nbctl in daemon mode to improve performance in large clusters",
+		Destination: &NbctlDaemonMode,
+	},
+	cli.BoolFlag{
+		Name:        "unprivileged-mode",
+		Usage:       "Run ovnkube-node container in unprivileged mode. Valid only with --init-node option.",
+		Destination: &UnprivilegedMode,
+	},
+	cli.BoolFlag{
+		Name:        "enable-multicast",
+		Usage:       "Adds multicast support. Valid only with --init-master option.",
+		Destination: &EnableMulticast,
+	},
 	// Logging options
 	cli.IntFlag{
 		Name:        "loglevel",
 		Usage:       "log verbosity and level: 5=debug, 4=info, 3=warn, 2=error, 1=fatal (default: 4)",
 		Destination: &cliConfig.Logging.Level,
+		Value:       Logging.Level,
 	},
 	cli.StringFlag{
 		Name:        "logfile",
@@ -397,11 +478,13 @@ var CNIFlags = []cli.Flag{
 		Name:        "cni-conf-dir",
 		Usage:       "the CNI config directory in which to write the overlay CNI config file (default: /etc/cni/net.d)",
 		Destination: &cliConfig.CNI.ConfDir,
+		Value:       CNI.ConfDir,
 	},
 	cli.StringFlag{
 		Name:        "cni-plugin",
 		Usage:       "the name of the CNI plugin (default: ovn-k8s-cni-overlay)",
 		Destination: &cliConfig.CNI.Plugin,
+		Value:       CNI.Plugin,
 	},
 	cli.StringFlag{
 		Name:        "win-hnsnetwork-id",
@@ -424,6 +507,7 @@ var K8sFlags = []cli.Flag{
 			"provided for kube-apiserver \"-service-cluster-ip-range\" " +
 			"option. (default: 172.16.1.0/24)",
 		Destination: &cliConfig.Kubernetes.ServiceCIDR,
+		Value:       Kubernetes.ServiceCIDR,
 	},
 	cli.StringFlag{
 		Name:        "k8s-kubeconfig",
@@ -434,6 +518,7 @@ var K8sFlags = []cli.Flag{
 		Name:        "k8s-apiserver",
 		Usage:       "URL of the Kubernetes API server (not required if --k8s-kubeconfig is given) (default: http://localhost:8443)",
 		Destination: &cliConfig.Kubernetes.APIServer,
+		Value:       Kubernetes.APIServer,
 	},
 	cli.StringFlag{
 		Name:        "k8s-cacert",
@@ -449,6 +534,35 @@ var K8sFlags = []cli.Flag{
 		Name:        "ovn-config-namespace",
 		Usage:       "specify a namespace which will contain services to config the OVN databases",
 		Destination: &cliConfig.Kubernetes.OVNConfigNamespace,
+		Value:       Kubernetes.OVNConfigNamespace,
+	},
+	cli.StringFlag{
+		Name:        "metrics-bind-address",
+		Usage:       "The IP address and port for the metrics server to serve on (set to 0.0.0.0 for all IPv4 interfaces)",
+		Destination: &cliConfig.Kubernetes.MetricsBindAddress,
+	},
+	cli.BoolFlag{
+		Name:        "metrics-enable-pprof",
+		Usage:       "If true, then also accept pprof requests on the metrics port.",
+		Destination: &cliConfig.Kubernetes.MetricsEnablePprof,
+	},
+	cli.BoolFlag{
+		Name: "ovn-empty-lb-events",
+		Usage: "If set, then load balancers do not get deleted when all backends are removed. " +
+			"Instead, ovn-kubernetes monitors the OVN southbound database for empty lb backends " +
+			"controller events. If one arrives, then a NeedPods event is sent so that Kubernetes " +
+			"will spin up pods for the load balancer to send traffic to.",
+		Destination: &cliConfig.Kubernetes.OVNEmptyLbEvents,
+	},
+	cli.StringFlag{
+		Name:        "pod-ip",
+		Usage:       "specify the ovnkube pod IP.",
+		Destination: &cliConfig.Kubernetes.PodIP,
+	},
+	cli.StringFlag{
+		Name:        "no-hostsubnet-nodes",
+		Usage:       "Specify a label for nodes that will manage their own hostsubnets",
+		Destination: &cliConfig.Kubernetes.RawNoHostSubnetNodes,
 	},
 }
 
@@ -462,18 +576,21 @@ var OvnNBFlags = []cli.Flag{
 		Destination: &cliConfig.OvnNorth.Address,
 	},
 	cli.StringFlag{
-		Name:        "nb-client-privkey",
-		Usage:       "Private key that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnnb-privkey.pem)",
+		Name: "nb-client-privkey",
+		Usage: "Private key that the client should use for talking to the OVN database (default when ssl address is used: /etc/openvswitch/ovnnb-privkey.pem).  " +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnNorth.PrivKey,
 	},
 	cli.StringFlag{
-		Name:        "nb-client-cert",
-		Usage:       "Client certificate that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnnb-cert.pem)",
+		Name: "nb-client-cert",
+		Usage: "Client certificate that the client should use for talking to the OVN database (default when ssl address is used: /etc/openvswitch/ovnnb-cert.pem). " +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnNorth.Cert,
 	},
 	cli.StringFlag{
-		Name:        "nb-client-cacert",
-		Usage:       "CA certificate that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnnb-ca.cert)",
+		Name: "nb-client-cacert",
+		Usage: "CA certificate that the client should use for talking to the OVN database (default when ssl address is used: /etc/openvswitch/ovnnb-ca.cert)." +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnNorth.CACert,
 	},
 }
@@ -488,18 +605,21 @@ var OvnSBFlags = []cli.Flag{
 		Destination: &cliConfig.OvnSouth.Address,
 	},
 	cli.StringFlag{
-		Name:        "sb-client-privkey",
-		Usage:       "Private key that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnsb-privkey.pem)",
+		Name: "sb-client-privkey",
+		Usage: "Private key that the client should use for talking to the OVN database (default when ssl address is used: /etc/openvswitch/ovnsb-privkey.pem)." +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnSouth.PrivKey,
 	},
 	cli.StringFlag{
-		Name:        "sb-client-cert",
-		Usage:       "Client certificate that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnsb-cert.pem)",
+		Name: "sb-client-cert",
+		Usage: "Client certificate that the client should use for talking to the OVN database(default when ssl address is used: /etc/openvswitch/ovnsb-cert.pem).  " +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnSouth.Cert,
 	},
 	cli.StringFlag{
-		Name:        "sb-client-cacert",
-		Usage:       "CA certificate that the client should use for talking to the OVN database.  Leave empty to use local unix socket. (default: /etc/openvswitch/ovnsb-ca.cert)",
+		Name: "sb-client-cacert",
+		Usage: "CA certificate that the client should use for talking to the OVN database (default when ssl address is used /etc/openvswitch/ovnsb-ca.cert). " +
+			"Default value for this setting is empty which defaults to use local unix socket.",
 		Destination: &cliConfig.OvnSouth.CACert,
 	},
 }
@@ -508,8 +628,8 @@ var OvnSBFlags = []cli.Flag{
 var OVNGatewayFlags = []cli.Flag{
 	cli.StringFlag{
 		Name: "gateway-mode",
-		Usage: "Sets the cluster gateway mode. One of \"shared\", \"spare\", " +
-			"or \"local\". If not given gateway functionality is disabled.",
+		Usage: "Sets the cluster gateway mode. One of \"shared\", " +
+			"or \"local\". If not given, gateway functionality is disabled.",
 	},
 	cli.StringFlag{
 		Name: "gateway-interface",
@@ -531,7 +651,7 @@ var OVNGatewayFlags = []cli.Flag{
 	cli.UintFlag{
 		Name: "gateway-vlanid",
 		Usage: "The VLAN on which the external network is available. " +
-			"Valid only for Shared or Spare Gateway interface mode.",
+			"Valid only for Shared Gateway interface mode.",
 		Destination: &cliConfig.Gateway.VLANID,
 	},
 	cli.BoolFlag{
@@ -546,11 +666,6 @@ var OVNGatewayFlags = []cli.Flag{
 		Usage:       "DEPRECATED; use --gateway-mode instead",
 		Destination: &initGateways,
 	},
-	cli.BoolFlag{
-		Name:        "gateway-spare-interface",
-		Usage:       "DEPRECATED; use --gateway-mode instead",
-		Destination: &gatewaySpareInterface,
-	},
 	cli.BoolFlag{
 		Name:        "gateway-local",
 		Usage:       "DEPRECATED; use --gateway-mode instead",
@@ -558,6 +673,45 @@ var OVNGatewayFlags = []cli.Flag{
 	},
 }
 
+// MasterHAFlags capture OVN northbound database options
+var MasterHAFlags = []cli.Flag{
+	cli.BoolFlag{
+		Name:        "manage-db-servers",
+		Usage:       "Manages the OVN North and South DB servers in active/passive",
+		Destination: &cliConfig.MasterHA.ManageDBServers,
+	},
+	cli.IntFlag{
+		Name:        "nb-port",
+		Usage:       "Port of the OVN northbound DB server to configure (default: 6641)",
+		Destination: &cliConfig.MasterHA.NbPort,
+		Value:       MasterHA.NbPort,
+	},
+	cli.IntFlag{
+		Name:        "sb-port",
+		Usage:       "Port of the OVN southbound DB server to configure (default: 6642)",
+		Destination: &cliConfig.MasterHA.SbPort,
+		Value:       MasterHA.SbPort,
+	},
+	cli.IntFlag{
+		Name:        "ha-election-lease-duration",
+		Usage:       "Leader election lease duration (in secs) (default: 60)",
+		Destination: &cliConfig.MasterHA.ElectionLeaseDuration,
+		Value:       MasterHA.ElectionLeaseDuration,
+	},
+	cli.IntFlag{
+		Name:        "ha-election-renew-deadline",
+		Usage:       "Leader election renew deadline (in secs) (default: 35)",
+		Destination: &cliConfig.MasterHA.ElectionRenewDeadline,
+		Value:       MasterHA.ElectionRenewDeadline,
+	},
+	cli.IntFlag{
+		Name:        "ha-election-retry-period",
+		Usage:       "Leader election retry period (in secs) (default: 10)",
+		Destination: &cliConfig.MasterHA.ElectionRetryPeriod,
+		Value:       MasterHA.ElectionRetryPeriod,
+	},
+}
+
 // Flags are general command-line flags. Apps should add these flags to their
 // own urfave/cli flags and call InitConfig() early in the application.
 var Flags []cli.Flag
@@ -630,14 +784,48 @@ func setOVSExternalID(exec kexec.Interface, key, value string) error {
 
 func buildKubernetesConfig(exec kexec.Interface, cli, file *config, saPath string, defaults *Defaults) error {
 	// token adn ca.crt may be from files mounted in container.
-	var saConfig KubernetesConfig
+	saConfig := savedKubernetes
 	if data, err := ioutil.ReadFile(filepath.Join(saPath, kubeServiceAccountFileToken)); err == nil {
 		saConfig.Token = string(data)
 	}
 	if _, err2 := os.Stat(filepath.Join(saPath, kubeServiceAccountFileCACert)); err2 == nil {
 		saConfig.CACert = filepath.Join(saPath, kubeServiceAccountFileCACert)
 	}
-	overrideFields(&Kubernetes, &saConfig)
+
+	if err := overrideFields(&Kubernetes, &saConfig, &savedKubernetes); err != nil {
+		return err
+	}
+
+	// values for token, cacert, kubeconfig, api-server may be found in several places.
+	// Priority order (highest first): OVS config, command line options, config file,
+	// environment variables, service account files
+
+	envConfig := savedKubernetes
+	envVarsMap := map[string]string{
+		"Kubeconfig": "KUBECONFIG",
+		"CACert":     "K8S_CACERT",
+		"APIServer":  "K8S_APISERVER",
+		"Token":      "K8S_TOKEN",
+	}
+	for k, v := range envVarsMap {
+		if x, exists := os.LookupEnv(v); exists && len(x) > 0 {
+			reflect.ValueOf(&envConfig).Elem().FieldByName(k).SetString(x)
+		}
+	}
+
+	if err := overrideFields(&Kubernetes, &envConfig, &savedKubernetes); err != nil {
+		return err
+	}
+
+	// Copy config file values over default values
+	if err := overrideFields(&Kubernetes, &file.Kubernetes, &savedKubernetes); err != nil {
+		return err
+	}
+
+	// And CLI overrides over config file and default values
+	if err := overrideFields(&Kubernetes, &cli.Kubernetes, &savedKubernetes); err != nil {
+		return err
+	}
 
 	// Grab default values from OVS external IDs
 	if defaults.K8sAPIServer {
@@ -650,23 +838,6 @@ func buildKubernetesConfig(exec kexec.Interface, cli, file *config, saPath strin
 		Kubernetes.CACert = getOVSExternalID(exec, "k8s-ca-certificate")
 	}
 
-	// values for token, cacert, kubeconfig, api-server may be found in several places.
-	// Take the first found when looking in this order: command line options, config file,
-	// environment variables, service account files
-
-	envConfig := KubernetesConfig{
-		Kubeconfig: os.Getenv("KUBECONFIG"),
-		CACert:     os.Getenv("K8S_CACERT"),
-		APIServer:  os.Getenv("K8S_APISERVER"),
-		Token:      os.Getenv("K8S_TOKEN"),
-	}
-	overrideFields(&Kubernetes, &envConfig)
-
-	// Copy config file values over default values
-	overrideFields(&Kubernetes, &file.Kubernetes)
-	// And CLI overrides over config file and default values
-	overrideFields(&Kubernetes, &cli.Kubernetes)
-
 	if Kubernetes.Kubeconfig != "" && !pathExists(Kubernetes.Kubeconfig) {
 		return fmt.Errorf("kubernetes kubeconfig file %q not found", Kubernetes.Kubeconfig)
 	}
@@ -691,30 +862,45 @@ func buildKubernetesConfig(exec kexec.Interface, cli, file *config, saPath strin
 		return fmt.Errorf("kubernetes service network CIDR %q invalid: %v", Kubernetes.ServiceCIDR, err)
 	}
 
+	if Kubernetes.PodIP != "" {
+		if ip := net.ParseIP(Kubernetes.PodIP); ip == nil {
+			return fmt.Errorf("Pod IP is invalid")
+		}
+	}
+
+	if Kubernetes.RawNoHostSubnetNodes != "" {
+		if nodeSelector, err := metav1.ParseToLabelSelector(Kubernetes.RawNoHostSubnetNodes); err == nil {
+			Kubernetes.NoHostSubnetNodes = nodeSelector
+		} else {
+			return fmt.Errorf("labelSelector \"%s\" is invalid: %v", Kubernetes.RawNoHostSubnetNodes, err)
+		}
+	}
 	return nil
 }
 
 func buildGatewayConfig(ctx *cli.Context, cli, file *config) error {
 	// Copy config file values over default values
-	overrideFields(&Gateway, &file.Gateway)
+	if err := overrideFields(&Gateway, &file.Gateway, &savedGateway); err != nil {
+		return err
+	}
 
 	cli.Gateway.Mode = GatewayMode(ctx.String("gateway-mode"))
 	if cli.Gateway.Mode == GatewayModeDisabled {
 		// Handle legacy CLI options
 		if ctx.Bool("init-gateways") {
 			cli.Gateway.Mode = GatewayModeShared
-			if ctx.Bool("gateway-spare-interface") {
-				cli.Gateway.Mode = GatewayModeSpare
-			} else if ctx.Bool("gateway-local") {
+			if ctx.Bool("gateway-local") {
 				cli.Gateway.Mode = GatewayModeLocal
 			}
 		}
 	}
 	// And CLI overrides over config file and default values
-	overrideFields(&Gateway, &cli.Gateway)
+	if err := overrideFields(&Gateway, &cli.Gateway, &savedGateway); err != nil {
+		return err
+	}
 
 	if Gateway.Mode != GatewayModeDisabled {
-		validModes := []string{string(GatewayModeShared), string(GatewayModeSpare), string(GatewayModeLocal)}
+		validModes := []string{string(GatewayModeShared), string(GatewayModeLocal)}
 		var found bool
 		for _, mode := range validModes {
 			if string(Gateway.Mode) == mode {
@@ -742,9 +928,39 @@ func buildGatewayConfig(ctx *cli.Context, cli, file *config) error {
 	return nil
 }
 
+func buildMasterHAConfig(ctx *cli.Context, cli, file *config) error {
+	// Copy config file values over default values
+	if err := overrideFields(&MasterHA, &file.MasterHA, &savedMasterHA); err != nil {
+		return err
+	}
+
+	// And CLI overrides over config file and default values
+	if err := overrideFields(&MasterHA, &cli.MasterHA, &savedMasterHA); err != nil {
+		return err
+	}
+
+	if MasterHA.ElectionLeaseDuration <= MasterHA.ElectionRenewDeadline {
+		return fmt.Errorf("Invalid HA election lease duration '%d'. "+
+			"It should be greater than HA election renew deadline '%d'",
+			MasterHA.ElectionLeaseDuration, MasterHA.ElectionRenewDeadline)
+	}
+
+	if MasterHA.ElectionRenewDeadline <= MasterHA.ElectionRetryPeriod {
+		return fmt.Errorf("Invalid HA election renew deadline duration '%d'. "+
+			"It should be greater than HA election retry period '%d'",
+			MasterHA.ElectionRenewDeadline, MasterHA.ElectionRetryPeriod)
+	}
+	return nil
+}
+
 func buildDefaultConfig(cli, file *config) error {
-	overrideFields(&Default, &file.Default)
-	overrideFields(&Default, &cli.Default)
+	if err := overrideFields(&Default, &file.Default, &savedDefault); err != nil {
+		return err
+	}
+
+	if err := overrideFields(&Default, &cli.Default, &savedDefault); err != nil {
+		return err
+	}
 
 	// Legacy cluster-subnet CLI option overrides config file or --cluster-subnets
 	if clusterSubnet != "" {
@@ -799,10 +1015,21 @@ func InitConfigSa(ctx *cli.Context, exec kexec.Interface, saPath string, default
 // common command-line options and constructs the global config object from
 // them. It returns the config file path (if explicitly specified) or an error
 func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, defaults *Defaults) (string, error) {
-	var cfg config
 	var retConfigFile string
 	var configFile string
 	var configFileIsDefault bool
+	var err error
+	// initialize cfg with default values, allow file read to override
+	cfg := config{
+		Default:    savedDefault,
+		Logging:    savedLogging,
+		CNI:        savedCNI,
+		Kubernetes: savedKubernetes,
+		OvnNorth:   savedOvnNorth,
+		OvnSouth:   savedOvnSouth,
+		Gateway:    savedGateway,
+		MasterHA:   savedMasterHA,
+	}
 
 	configFile, configFileIsDefault = getConfigFilePath(ctx)
 
@@ -834,27 +1061,30 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 	}
 
 	// Build config that needs no special processing
-	overrideFields(&CNI, &cfg.CNI)
-	overrideFields(&CNI, &cliConfig.CNI)
+	if err = overrideFields(&CNI, &cfg.CNI, &savedCNI); err != nil {
+		return "", err
+	}
+	if err = overrideFields(&CNI, &cliConfig.CNI, &savedCNI); err != nil {
+		return "", err
+	}
 
 	// Logging setup
-	overrideFields(&Logging, &cfg.Logging)
-	overrideFields(&Logging, &cliConfig.Logging)
+	if err = overrideFields(&Logging, &cfg.Logging, &savedLogging); err != nil {
+		return "", err
+	}
+	if err = overrideFields(&Logging, &cliConfig.Logging, &savedLogging); err != nil {
+		return "", err
+	}
+
 	logrus.SetLevel(logrus.Level(Logging.Level))
 	if Logging.File != "" {
-		var file *os.File
-		if _, err = os.Stat(filepath.Dir(Logging.File)); os.IsNotExist(err) {
-			dir := filepath.Dir(Logging.File)
-			if err = os.MkdirAll(dir, 0755); err != nil {
-				logrus.Errorf("failed to create logfile directory %s (%v). Ignoring..", dir, err)
-			}
-		}
-		file, err = os.OpenFile(Logging.File, os.O_CREATE|os.O_APPEND|os.O_WRONLY, 0660)
-		if err != nil {
-			logrus.Errorf("failed to open logfile %s (%v). Ignoring..", Logging.File, err)
-		} else {
-			logrus.SetOutput(file)
-		}
+		logrus.SetOutput(&lumberjack.Logger{
+			Filename:   Logging.File,
+			MaxSize:    100, // megabytes
+			MaxBackups: 10,
+			MaxAge:     30, // days
+			Compress:   true,
+		})
 	}
 
 	if err = buildDefaultConfig(&cliConfig, &cfg); err != nil {
@@ -869,6 +1099,10 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 		return "", err
 	}
 
+	if err = buildMasterHAConfig(ctx, &cliConfig, &cfg); err != nil {
+		return "", err
+	}
+
 	tmpAuth, err := buildOvnAuth(exec, true, &cliConfig.OvnNorth, &cfg.OvnNorth, defaults.OvnNorthAddress)
 	if err != nil {
 		return "", err
@@ -881,6 +1115,12 @@ func initConfigWithPath(ctx *cli.Context, exec kexec.Interface, saPath string, d
 	}
 	OvnSouth = *tmpAuth
 
+	// Determine if ovn-kubernetes is configured to run in IPv6 mode
+	IPv6Mode = false
+	if len(Default.ClusterSubnets) >= 1 && Default.ClusterSubnets[0].CIDR.IP.To4() == nil {
+		IPv6Mode = true
+	}
+
 	logrus.Debugf("Default config: %+v", Default)
 	logrus.Debugf("Logging config: %+v", Logging)
 	logrus.Debugf("CNI config: %+v", CNI)
@@ -901,6 +1141,7 @@ func pathExists(path string) bool {
 
 // parseAddress parses an OVN database address, which can be of form
 // "ssl:1.2.3.4:6641,ssl:1.2.3.5:6641" or "ssl://1.2.3.4:6641,ssl://1.2.3.5:6641"
+// or "ssl:[fd01::1]:6641,ssl:[fd01::2]:6641
 // and returns the validated address(es) and the scheme
 func parseAddress(urlString string) (string, OvnDBScheme, error) {
 	var parsedAddress, scheme string
@@ -908,11 +1149,10 @@ func parseAddress(urlString string) (string, OvnDBScheme, error) {
 
 	urlString = strings.Replace(urlString, "//", "", -1)
 	for _, ovnAddress := range strings.Split(urlString, ",") {
-		splits := strings.Split(ovnAddress, ":")
-		if len(splits) != 3 {
+		splits := strings.SplitN(ovnAddress, ":", 2)
+		if len(splits) != 2 {
 			return "", "", fmt.Errorf("Failed to parse OVN address %s", urlString)
 		}
-		hostPort := splits[1] + ":" + splits[2]
 
 		if scheme == "" {
 			scheme = splits[0]
@@ -921,21 +1161,16 @@ func parseAddress(urlString string) (string, OvnDBScheme, error) {
 				urlString)
 		}
 
-		host, port, err := net.SplitHostPort(hostPort)
+		host, port, err := net.SplitHostPort(splits[1])
 		if err != nil {
 			return "", "", fmt.Errorf("failed to parse OVN DB host/port %q: %v",
-				hostPort, err)
-		}
-		ip := net.ParseIP(host)
-		if ip == nil {
-			return "", "", fmt.Errorf("OVN DB host %q must be an IP address, "+
-				"not a DNS name", hostPort)
+				splits[1], err)
 		}
 
 		if parsedAddress != "" {
 			parsedAddress += ","
 		}
-		parsedAddress += fmt.Sprintf("%s:%s:%s", scheme, host, port)
+		parsedAddress += fmt.Sprintf("%s:%s", scheme, net.JoinHostPort(host, port))
 	}
 
 	switch {
@@ -959,12 +1194,15 @@ func buildOvnAuth(exec kexec.Interface, northbound bool, cliAuth, confAuth *OvnA
 	}
 
 	var direction string
+	var defaultAuth *OvnAuthConfig
 	if northbound {
 		auth.externalID = "ovn-nb"
 		direction = "nb"
+		defaultAuth = &savedOvnNorth
 	} else {
 		auth.externalID = "ovn-remote"
 		direction = "sb"
+		defaultAuth = &savedOvnSouth
 	}
 
 	// Determine final address so we know how to set cert/key defaults
@@ -981,9 +1219,28 @@ func buildOvnAuth(exec kexec.Interface, northbound bool, cliAuth, confAuth *OvnA
 		auth.PrivKey = "/etc/openvswitch/ovn" + direction + "-privkey.pem"
 		auth.Cert = "/etc/openvswitch/ovn" + direction + "-cert.pem"
 	}
+
 	// Build the final auth config with overrides from CLI and config file
-	overrideFields(auth, confAuth)
-	overrideFields(auth, cliAuth)
+	if err := overrideFields(auth, confAuth, defaultAuth); err != nil {
+		return nil, err
+	}
+	if err := overrideFields(auth, cliAuth, defaultAuth); err != nil {
+		return nil, err
+	}
+
+	// When instructed to watch-endpoint, just set scheme and obtain address(es) from endpoint
+	if address == "watch-endpoint" {
+		if !MasterHA.ManageDBServers {
+			return nil, fmt.Errorf("watch-endpoint requires --manage-db-servers")
+		}
+		// Set scheme and wait for an endpoint update to provide address/port to use
+		if auth.PrivKey != "" || auth.Cert != "" || auth.CACert != "" {
+			auth.Scheme = OvnDBSchemeSSL
+		} else {
+			auth.Scheme = OvnDBSchemeTCP
+		}
+		return auth, nil
+	}
 
 	if address == "" {
 		if auth.PrivKey != "" || auth.Cert != "" || auth.CACert != "" {
@@ -991,6 +1248,11 @@ func buildOvnAuth(exec kexec.Interface, northbound bool, cliAuth, confAuth *OvnA
 		}
 		auth.Scheme = OvnDBSchemeUnix
 		return auth, nil
+	} else if MasterHA.ManageDBServers {
+		if northbound {
+			return nil, fmt.Errorf("--nb-address is not allowed with --manage-db-servers")
+		}
+		return nil, fmt.Errorf("--sb-address is not allowed with --manage-db-servers")
 	}
 
 	var err error
@@ -1093,38 +1355,18 @@ func (a *OvnAuthConfig) SetDBAuth() error {
 	return nil
 }
 
-func (a *OvnAuthConfig) updateIP(newIP []string, port string) error {
-	if a.Address != "" {
-		s := strings.Split(a.Address, ":")
-		if len(s) != 3 {
-			return fmt.Errorf("failed to parse OvnAuthConfig address %q", a.Address)
-		}
-		var newPort string
-		if port != "" {
-			newPort = port
-		} else {
-			newPort = s[2]
-		}
-
-		newAddresses := make([]string, 0, len(newIP))
-		for _, ipAddress := range newIP {
-			newAddresses = append(newAddresses, s[0]+":"+ipAddress+":"+newPort)
-		}
-		a.Address = strings.Join(newAddresses, ",")
+func (a *OvnAuthConfig) updateIP(newIPs []string, port string) {
+	newAddresses := make([]string, 0, len(newIPs))
+	for _, ipAddress := range newIPs {
+		newAddresses = append(newAddresses, fmt.Sprintf("%v:%s", a.Scheme, net.JoinHostPort(ipAddress, port)))
 	}
-	return nil
+	a.Address = strings.Join(newAddresses, ",")
 }
 
 // UpdateOVNNodeAuth updates the host and URL in ClientAuth
 // for both OvnNorth and OvnSouth. It updates them with the new masterIP.
-func UpdateOVNNodeAuth(masterIP []string, southboundDBPort, northboundDBPort string) error {
+func UpdateOVNNodeAuth(masterIP []string, southboundDBPort, northboundDBPort string) {
 	logrus.Debugf("Update OVN node auth with new master ip: %s", masterIP)
-	if err := OvnNorth.updateIP(masterIP, northboundDBPort); err != nil {
-		return fmt.Errorf("failed to update OvnNorth ClientAuth URL: %v", err)
-	}
-
-	if err := OvnSouth.updateIP(masterIP, southboundDBPort); err != nil {
-		return fmt.Errorf("failed to update OvnSouth ClientAuth URL: %v", err)
-	}
-	return nil
+	OvnNorth.updateIP(masterIP, northboundDBPort)
+	OvnSouth.updateIP(masterIP, southboundDBPort)
 }
