@@ -1,7 +1,10 @@
 package factory
 
 import (
+	"fmt"
 	"reflect"
+	"sync"
+	"sync/atomic"
 	"testing"
 
 	"k8s.io/api/core/v1"
@@ -102,6 +105,24 @@ func objSetup(c *fake.Clientset, objType string, listFn func(core.Action) (bool,
 	return w
 }
 
+type handlerCalls struct {
+	added   int32
+	updated int32
+	deleted int32
+}
+
+func (c *handlerCalls) getAdded() int {
+	return int(atomic.LoadInt32(&c.added))
+}
+
+func (c *handlerCalls) getUpdated() int {
+	return int(atomic.LoadInt32(&c.updated))
+}
+
+func (c *handlerCalls) getDeleted() int {
+	return int(atomic.LoadInt32(&c.deleted))
+}
+
 var _ = Describe("Watch Factory Operations", func() {
 	var (
 		fakeClient                                *fake.Clientset
@@ -114,7 +135,6 @@ var _ = Describe("Watch Factory Operations", func() {
 		endpoints                                 []*v1.Endpoints
 		services                                  []*v1.Service
 		stop                                      chan struct{}
-		numAdded, numUpdated, numDeleted          int
 	)
 
 	BeforeEach(func() {
@@ -174,10 +194,10 @@ var _ = Describe("Watch Factory Operations", func() {
 			}
 			return true, obj, nil
 		})
+	})
 
-		numAdded = 0
-		numUpdated = 0
-		numDeleted = 0
+	AfterEach(func() {
+		close(stop)
 	})
 
 	Context("when a processExisting is given", func() {
@@ -187,12 +207,12 @@ var _ = Describe("Watch Factory Operations", func() {
 			h, err := wf.addHandler(objType, namespace, lsel,
 				cache.ResourceEventHandlerFuncs{},
 				func(objs []interface{}) {
+					defer GinkgoRecover()
 					Expect(len(objs)).To(Equal(1))
 				})
 			Expect(err).NotTo(HaveOccurred())
 			Expect(h).NotTo(BeNil())
 			wf.removeHandler(objType, h)
-			wf.Shutdown()
 		}
 
 		It("is called for each existing pod", func() {
@@ -239,18 +259,18 @@ var _ = Describe("Watch Factory Operations", func() {
 		testExisting := func(objType reflect.Type) {
 			wf, err := NewWatchFactory(fakeClient, stop)
 			Expect(err).NotTo(HaveOccurred())
+			var addCalls int32
 			h, err := wf.addHandler(objType, "", nil,
 				cache.ResourceEventHandlerFuncs{
 					AddFunc: func(obj interface{}) {
-						numAdded++
+						atomic.AddInt32(&addCalls, 1)
 					},
 					UpdateFunc: func(old, new interface{}) {},
 					DeleteFunc: func(obj interface{}) {},
 				}, nil)
 			Expect(err).NotTo(HaveOccurred())
-			Expect(numAdded).To(Equal(2))
+			Expect(int(addCalls)).To(Equal(2))
 			wf.removeHandler(objType, h)
-			wf.Shutdown()
 		}
 
 		It("calls ADD for each existing pod", func() {
@@ -290,30 +310,31 @@ var _ = Describe("Watch Factory Operations", func() {
 		})
 	})
 
-	addFilteredHandler := func(wf *WatchFactory, objType reflect.Type, namespace string, lsel *metav1.LabelSelector, funcs cache.ResourceEventHandlerFuncs) *Handler {
+	addFilteredHandler := func(wf *WatchFactory, objType reflect.Type, namespace string, lsel *metav1.LabelSelector, funcs cache.ResourceEventHandlerFuncs) (*Handler, *handlerCalls) {
+		calls := handlerCalls{}
 		h, err := wf.addHandler(objType, namespace, lsel, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				defer GinkgoRecover()
-				numAdded++
+				atomic.AddInt32(&calls.added, 1)
 				funcs.AddFunc(obj)
 			},
 			UpdateFunc: func(old, new interface{}) {
 				defer GinkgoRecover()
-				numUpdated++
+				atomic.AddInt32(&calls.updated, 1)
 				funcs.UpdateFunc(old, new)
 			},
 			DeleteFunc: func(obj interface{}) {
 				defer GinkgoRecover()
-				numDeleted++
+				atomic.AddInt32(&calls.deleted, 1)
 				funcs.DeleteFunc(obj)
 			},
 		}, nil)
 		Expect(err).NotTo(HaveOccurred())
 		Expect(h).NotTo(BeNil())
-		return h
+		return h, &calls
 	}
 
-	addHandler := func(wf *WatchFactory, objType reflect.Type, funcs cache.ResourceEventHandlerFuncs) *Handler {
+	addHandler := func(wf *WatchFactory, objType reflect.Type, funcs cache.ResourceEventHandlerFuncs) (*Handler, *handlerCalls) {
 		return addFilteredHandler(wf, objType, "", nil, funcs)
 	}
 
@@ -322,7 +343,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newPod("pod1", "default")
-		h := addHandler(wf, podType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, podType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				pod := obj.(*v1.Pod)
 				Expect(reflect.DeepEqual(pod, added)).To(BeTrue())
@@ -340,16 +361,85 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		pods = append(pods, added)
 		podWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Spec.NodeName = "foobar"
 		podWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		pods = pods[:0]
 		podWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
+
+		wf.RemovePodHandler(h)
+	})
+
+	It("responds to multiple pod add/update/delete events", func() {
+		wf, err := NewWatchFactory(fakeClient, stop)
+		Expect(err).NotTo(HaveOccurred())
+
+		const nodeName string = "mynode"
+		type opTest struct {
+			pod     *v1.Pod
+			added   int
+			updated int
+			deleted int
+		}
+		testPods := make(map[string]*opTest)
+
+		for i := 0; i < 5; i++ {
+			name := fmt.Sprintf("mypod-%d", i)
+			pod := newPod(name, fmt.Sprintf("namespace-%d", i))
+			testPods[name] = &opTest{pod: pod}
+		}
+
+		h, c := addHandler(wf, podType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				pod := obj.(*v1.Pod)
+				ot, ok := testPods[pod.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.added).To(BeNumerically("<", 2))
+				ot.added++
+			},
+			UpdateFunc: func(old, new interface{}) {
+				newPod := new.(*v1.Pod)
+				ot, ok := testPods[newPod.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.updated).To(BeNumerically("<", 2))
+				ot.updated++
+				Expect(newPod.Spec.NodeName).To(Equal(nodeName))
+			},
+			DeleteFunc: func(obj interface{}) {
+				pod := obj.(*v1.Pod)
+				ot, ok := testPods[pod.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.deleted).To(BeNumerically("<", 2))
+				ot.deleted++
+			},
+		})
+
+		// Add/Update/Delete each pod twice
+		for i := 0; i < 2; i++ {
+			for _, ot := range testPods {
+				pods = append(pods, ot.pod)
+				podWatch.Add(ot.pod)
+				ot.pod.Spec.NodeName = nodeName
+				podWatch.Modify(ot.pod)
+				pods = pods[:0]
+				podWatch.Delete(ot.pod)
+			}
+		}
+
+		// Ensure total number of each operation is 10; and each
+		// node's individual operation count is 2
+		Eventually(c.getAdded, 2).Should(Equal(10))
+		Eventually(c.getUpdated, 2).Should(Equal(10))
+		Eventually(c.getDeleted, 2).Should(Equal(10))
+		for _, ot := range testPods {
+			Expect(ot.added).Should(Equal(2))
+			Expect(ot.updated).Should(Equal(2))
+			Expect(ot.deleted).Should(Equal(2))
+		}
 
 		wf.RemovePodHandler(h)
-		wf.Shutdown()
 	})
 
 	It("responds to namespace add/update/delete events", func() {
@@ -357,7 +447,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNamespace("default")
-		h := addHandler(wf, namespaceType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, namespaceType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				ns := obj.(*v1.Namespace)
 				Expect(reflect.DeepEqual(ns, added)).To(BeTrue())
@@ -375,16 +465,15 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		namespaces = append(namespaces, added)
 		namespaceWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Status.Phase = v1.NamespaceTerminating
 		namespaceWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		namespaces = namespaces[:0]
 		namespaceWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 
 		wf.RemoveNamespaceHandler(h)
-		wf.Shutdown()
 	})
 
 	It("responds to node add/update/delete events", func() {
@@ -392,7 +481,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNode("mynode")
-		h := addHandler(wf, nodeType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, nodeType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				node := obj.(*v1.Node)
 				Expect(reflect.DeepEqual(node, added)).To(BeTrue())
@@ -410,16 +499,230 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		nodes = append(nodes, added)
 		nodeWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Status.Phase = v1.NodeTerminated
 		nodeWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		nodes = nodes[:0]
 		nodeWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 
-		wf.removeHandler(nodeType, h)
-		wf.Shutdown()
+		wf.RemoveNodeHandler(h)
+	})
+
+	It("responds to multiple node add/update/delete events", func() {
+		wf, err := NewWatchFactory(fakeClient, stop)
+		Expect(err).NotTo(HaveOccurred())
+
+		type opTest struct {
+			node    *v1.Node
+			added   int
+			updated int
+			deleted int
+		}
+		testNodes := make(map[string]*opTest)
+
+		for i := 0; i < 5; i++ {
+			name := fmt.Sprintf("mynode-%d", i)
+			node := newNode(name)
+			testNodes[name] = &opTest{node: node}
+		}
+
+		h, c := addHandler(wf, nodeType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				node := obj.(*v1.Node)
+				ot, ok := testNodes[node.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.added).To(BeNumerically("<", 2))
+				ot.added++
+			},
+			UpdateFunc: func(old, new interface{}) {
+				newNode := new.(*v1.Node)
+				ot, ok := testNodes[newNode.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.updated).To(BeNumerically("<", 2))
+				ot.updated++
+				Expect(newNode.Status.Phase).To(Equal(v1.NodeTerminated))
+			},
+			DeleteFunc: func(obj interface{}) {
+				node := obj.(*v1.Node)
+				ot, ok := testNodes[node.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.deleted).To(BeNumerically("<", 2))
+				ot.deleted++
+			},
+		})
+
+		// Add/Update/Delete each node twice
+		for i := 0; i < 2; i++ {
+			for _, ot := range testNodes {
+				nodes = append(nodes, ot.node)
+				nodeWatch.Add(ot.node)
+				ot.node.Status.Phase = v1.NodeTerminated
+				nodeWatch.Modify(ot.node)
+				nodes = nodes[:0]
+				nodeWatch.Delete(ot.node)
+			}
+		}
+
+		// Ensure total number of each operation is 10; and each
+		// node's individual operation count is 2
+		Eventually(c.getAdded, 2).Should(Equal(10))
+		Eventually(c.getUpdated, 2).Should(Equal(10))
+		Eventually(c.getDeleted, 2).Should(Equal(10))
+		for _, ot := range testNodes {
+			Expect(ot.added).Should(Equal(2))
+			Expect(ot.updated).Should(Equal(2))
+			Expect(ot.deleted).Should(Equal(2))
+		}
+
+		wf.RemoveNodeHandler(h)
+	})
+
+	It("correctly orders queued informer initial add events and subsequent update events", func() {
+		type opTest struct {
+			node    *v1.Node
+			added   int
+			updated int
+		}
+		testNodes := make(map[string]*opTest)
+
+		for i := 0; i < 600; i++ {
+			name := fmt.Sprintf("mynode-%d", i)
+			node := newNode(name)
+			testNodes[name] = &opTest{node: node}
+			// Add all nodes to the initial list
+			nodes = append(nodes, node)
+		}
+
+		wf, err := NewWatchFactory(fakeClient, stop)
+		Expect(err).NotTo(HaveOccurred())
+
+		startWg := sync.WaitGroup{}
+		startWg.Add(1)
+		doneWg := sync.WaitGroup{}
+		doneWg.Add(1)
+		go func() {
+			startWg.Done()
+			// Send an update event for each node
+			for _, n := range nodes {
+				n.Status.Phase = v1.NodeTerminated
+				nodeWatch.Modify(n)
+			}
+			doneWg.Done()
+		}()
+		startWg.Wait()
+
+		h, c := addHandler(wf, nodeType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				defer GinkgoRecover()
+				node := obj.(*v1.Node)
+				ot, ok := testNodes[node.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.added).To(Equal(0), "add for node %s already run", node.Name)
+				ot.added++
+			},
+			UpdateFunc: func(old, new interface{}) {
+				defer GinkgoRecover()
+				newNode := new.(*v1.Node)
+				ot, ok := testNodes[newNode.Name]
+				Expect(ok).To(BeTrue())
+				// Expect updates to be processed after Add
+				Expect(ot.added).To(Equal(1), "update for node %s processed before initial add!", newNode.Name)
+				Expect(ot.updated).To(Equal(0))
+				ot.updated++
+				Expect(newNode.Status.Phase).To(Equal(v1.NodeTerminated))
+			},
+			DeleteFunc: func(obj interface{}) {},
+		})
+		doneWg.Wait()
+
+		// Adds are done synchronously at handler addition time
+		for _, ot := range testNodes {
+			Expect(ot.added).To(Equal(1), "missing add for node %s", ot.node.Name)
+		}
+		Expect(c.getAdded()).To(Equal(len(testNodes)))
+
+		// Updates are async and may take a bit longer to finish
+		Eventually(c.getUpdated, 10).Should(Equal(len(testNodes)))
+		for _, ot := range testNodes {
+			Expect(ot.updated).To(Equal(1), "missing update for node %s", ot.node.Name)
+		}
+
+		wf.RemoveNodeHandler(h)
+	})
+
+	It("correctly orders serialized informer initial add events and subsequent update events", func() {
+		type opTest struct {
+			namespace *v1.Namespace
+			added     int
+			updated   int
+		}
+		testNamespaces := make(map[string]*opTest)
+
+		for i := 0; i < 598; i++ {
+			name := fmt.Sprintf("mynamespace-%d", i)
+			namespace := newNamespace(name)
+			testNamespaces[name] = &opTest{namespace: namespace}
+			// Add all namespaces to the initial list
+			namespaces = append(namespaces, namespace)
+		}
+
+		wf, err := NewWatchFactory(fakeClient, stop)
+		Expect(err).NotTo(HaveOccurred())
+
+		startWg := sync.WaitGroup{}
+		startWg.Add(1)
+		doneWg := sync.WaitGroup{}
+		doneWg.Add(1)
+		go func() {
+			startWg.Done()
+			// Send an update event for each namespace
+			for _, n := range namespaces {
+				n.Status.Phase = v1.NamespaceTerminating
+				namespaceWatch.Modify(n)
+			}
+			doneWg.Done()
+		}()
+		startWg.Wait()
+
+		h, c := addHandler(wf, namespaceType, cache.ResourceEventHandlerFuncs{
+			AddFunc: func(obj interface{}) {
+				defer GinkgoRecover()
+				namespace := obj.(*v1.Namespace)
+				ot, ok := testNamespaces[namespace.Name]
+				Expect(ok).To(BeTrue())
+				Expect(ot.added).To(Equal(0))
+				ot.added++
+			},
+			UpdateFunc: func(old, new interface{}) {
+				defer GinkgoRecover()
+				newNamespace := new.(*v1.Namespace)
+				ot, ok := testNamespaces[newNamespace.Name]
+				Expect(ok).To(BeTrue())
+				// Expect updates to be processed after Add
+				Expect(ot.added).To(Equal(1), "update for namespace %s processed before initial add!", newNamespace.Name)
+				Expect(ot.updated).To(Equal(0))
+				ot.updated++
+				Expect(newNamespace.Status.Phase).To(Equal(v1.NamespaceTerminating))
+			},
+			DeleteFunc: func(obj interface{}) {},
+		})
+		doneWg.Wait()
+
+		// Adds are done synchronously at handler addition time
+		for _, ot := range testNamespaces {
+			Expect(ot.added).To(Equal(1), "missing add for namespace %s", ot.namespace.Name)
+		}
+		Expect(c.getAdded()).To(Equal(len(testNamespaces)))
+
+		// Updates are async and may take a bit longer to finish
+		Eventually(c.getUpdated, 10).Should(Equal(len(testNamespaces)))
+		for _, ot := range testNamespaces {
+			Expect(ot.updated).To(Equal(1), "missing update for namespace %s", ot.namespace.Name)
+		}
+
+		wf.RemoveNamespaceHandler(h)
 	})
 
 	It("responds to policy add/update/delete events", func() {
@@ -427,7 +730,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newPolicy("mypolicy", "default")
-		h := addHandler(wf, policyType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, policyType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				np := obj.(*knet.NetworkPolicy)
 				Expect(reflect.DeepEqual(np, added)).To(BeTrue())
@@ -445,16 +748,15 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		policies = append(policies, added)
 		policyWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Spec.PolicyTypes = []knet.PolicyType{knet.PolicyTypeIngress}
 		policyWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		policies = policies[:0]
 		policyWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 
-		wf.removeHandler(policyType, h)
-		wf.Shutdown()
+		wf.RemovePolicyHandler(h)
 	})
 
 	It("responds to endpoints add/update/delete events", func() {
@@ -462,7 +764,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newEndpoints("myendpoints", "default")
-		h := addHandler(wf, endpointsType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, endpointsType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				eps := obj.(*v1.Endpoints)
 				Expect(reflect.DeepEqual(eps, added)).To(BeTrue())
@@ -480,7 +782,7 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		endpoints = append(endpoints, added)
 		endpointsWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Subsets = append(added.Subsets, v1.EndpointSubset{
 			Ports: []v1.EndpointPort{
 				{
@@ -490,13 +792,12 @@ var _ = Describe("Watch Factory Operations", func() {
 			},
 		})
 		endpointsWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		endpoints = endpoints[:0]
 		endpointsWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 
-		wf.removeHandler(endpointsType, h)
-		wf.Shutdown()
+		wf.RemoveEndpointsHandler(h)
 	})
 
 	It("responds to service add/update/delete events", func() {
@@ -504,7 +805,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newService("myservice", "default")
-		h := addHandler(wf, serviceType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, serviceType, cache.ResourceEventHandlerFuncs{
 			AddFunc: func(obj interface{}) {
 				service := obj.(*v1.Service)
 				Expect(reflect.DeepEqual(service, added)).To(BeTrue())
@@ -522,16 +823,15 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		services = append(services, added)
 		serviceWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		added.Spec.ClusterIP = "1.1.1.1"
 		serviceWatch.Modify(added)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 		services = services[:0]
 		serviceWatch.Delete(added)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 
-		wf.removeHandler(serviceType, h)
-		wf.Shutdown()
+		wf.RemoveServiceHandler(h)
 	})
 
 	It("stops processing events after the handler is removed", func() {
@@ -539,7 +839,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		Expect(err).NotTo(HaveOccurred())
 
 		added := newNamespace("default")
-		h := addHandler(wf, namespaceType, cache.ResourceEventHandlerFuncs{
+		h, c := addHandler(wf, namespaceType, cache.ResourceEventHandlerFuncs{
 			AddFunc:    func(obj interface{}) {},
 			UpdateFunc: func(old, new interface{}) {},
 			DeleteFunc: func(obj interface{}) {},
@@ -547,22 +847,20 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		namespaces = append(namespaces, added)
 		namespaceWatch.Add(added)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 		wf.RemoveNamespaceHandler(h)
 
 		added2 := newNamespace("other")
 		namespaces = append(namespaces, added2)
 		namespaceWatch.Add(added2)
-		Consistently(func() int { return numAdded }, 2).Should(Equal(1))
+		Consistently(c.getAdded, 2).Should(Equal(1))
 
 		added2.Status.Phase = v1.NamespaceTerminating
 		namespaceWatch.Modify(added2)
-		Consistently(func() int { return numUpdated }, 2).Should(Equal(0))
+		Consistently(c.getUpdated, 2).Should(Equal(0))
 		namespaces = []*v1.Namespace{added}
 		namespaceWatch.Delete(added2)
-		Consistently(func() int { return numDeleted }, 2).Should(Equal(0))
-
-		wf.Shutdown()
+		Consistently(c.getDeleted, 2).Should(Equal(0))
 	})
 
 	It("filters correctly by label and namespace", func() {
@@ -576,7 +874,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		failsFilter2 := newPod("pod3", "otherns")
 		failsFilter2.ObjectMeta.Labels["blah"] = "foobar"
 
-		addFilteredHandler(wf,
+		_, c := addFilteredHandler(wf,
 			podType,
 			"default",
 			&metav1.LabelSelector{
@@ -599,36 +897,34 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		pods = append(pods, passesFilter)
 		podWatch.Add(passesFilter)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 
 		// numAdded should remain 1
 		pods = append(pods, failsFilter)
 		podWatch.Add(failsFilter)
-		Consistently(func() int { return numAdded }, 2).Should(Equal(1))
+		Consistently(c.getAdded, 2).Should(Equal(1))
 
 		// numAdded should remain 1
 		pods = append(pods, failsFilter2)
 		podWatch.Add(failsFilter2)
-		Consistently(func() int { return numAdded }, 2).Should(Equal(1))
+		Consistently(c.getAdded, 2).Should(Equal(1))
 
 		passesFilter.Status.Phase = v1.PodFailed
 		podWatch.Modify(passesFilter)
-		Eventually(func() int { return numUpdated }, 2).Should(Equal(1))
+		Eventually(c.getUpdated, 2).Should(Equal(1))
 
 		// numAdded should remain 1
 		failsFilter.Status.Phase = v1.PodFailed
 		podWatch.Modify(failsFilter)
-		Consistently(func() int { return numUpdated }, 2).Should(Equal(1))
+		Consistently(c.getUpdated, 2).Should(Equal(1))
 
 		failsFilter2.Status.Phase = v1.PodFailed
 		podWatch.Modify(failsFilter2)
-		Consistently(func() int { return numUpdated }, 2).Should(Equal(1))
+		Consistently(c.getUpdated, 2).Should(Equal(1))
 
 		pods = []*v1.Pod{failsFilter, failsFilter2}
 		podWatch.Delete(passesFilter)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
-
-		wf.Shutdown()
+		Eventually(c.getDeleted, 2).Should(Equal(1))
 	})
 
 	It("correctly handles object updates that cause filter changes", func() {
@@ -639,7 +935,7 @@ var _ = Describe("Watch Factory Operations", func() {
 		pod.ObjectMeta.Labels["blah"] = "baz"
 
 		equalPod := pod
-		h := addFilteredHandler(wf,
+		h, c := addFilteredHandler(wf,
 			podType,
 			"default",
 			&metav1.LabelSelector{
@@ -661,7 +957,7 @@ var _ = Describe("Watch Factory Operations", func() {
 
 		// Pod doesn't pass filter; shouldn't be added
 		podWatch.Add(pod)
-		Consistently(func() int { return numAdded }, 2).Should(Equal(0))
+		Consistently(c.getAdded, 2).Should(Equal(0))
 
 		// Update pod to pass filter; should be treated as add.  Need
 		// to deep-copy pod when modifying because it's a pointer all
@@ -671,16 +967,15 @@ var _ = Describe("Watch Factory Operations", func() {
 		pods = []*v1.Pod{podCopy}
 		equalPod = podCopy
 		podWatch.Modify(podCopy)
-		Eventually(func() int { return numAdded }, 2).Should(Equal(1))
+		Eventually(c.getAdded, 2).Should(Equal(1))
 
 		// Update pod to fail filter; should be treated as delete
 		pod.ObjectMeta.Labels["blah"] = "baz"
 		podWatch.Modify(pod)
-		Eventually(func() int { return numDeleted }, 2).Should(Equal(1))
-		Consistently(func() int { return numAdded }, 2).Should(Equal(1))
-		Consistently(func() int { return numUpdated }, 2).Should(Equal(0))
+		Eventually(c.getDeleted, 2).Should(Equal(1))
+		Consistently(c.getAdded, 2).Should(Equal(1))
+		Consistently(c.getUpdated, 2).Should(Equal(0))
 
 		wf.RemovePodHandler(h)
-		wf.Shutdown()
 	})
 })
