@@ -42,7 +42,7 @@ func (ovn *Controller) getLbEndpoints(ep *kapi.Endpoints, tcpPortMap, udpPortMap
 func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 	// get service
 	// TODO: cache the service
-	svc, err := ovn.kube.GetService(ep.Namespace, ep.Name)
+	svc, err := ovn.watchFactory.GetService(ep.Namespace, ep.Name)
 	if err != nil {
 		// This is not necessarily an error. For e.g when there are endpoints
 		// without a corresponding service.
@@ -63,7 +63,7 @@ func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 		targetPort := lbEps.Port
 		for _, svcPort := range svc.Spec.Ports {
 			if svcPort.Protocol == kapi.ProtocolTCP && svcPort.Name == svcPortName {
-				if util.ServiceTypeHasNodePort(svc) && config.Gateway.NodeportEnable {
+				if util.ServiceTypeHasNodePort(svc) {
 					logrus.Debugf("Creating Gateways IP for NodePort: %d, %v", svcPort.NodePort, ips)
 					err = ovn.createGatewaysVIP(string(svcPort.Protocol), svcPort.NodePort, targetPort, ips)
 					if err != nil {
@@ -85,6 +85,8 @@ func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 						logrus.Errorf("Error in creating Cluster IP for svc %s, target port: %d - %v\n", svc.Name, targetPort, err)
 						continue
 					}
+					vip := util.JoinHostPortInt32(svc.Spec.ClusterIP, svcPort.Port)
+					ovn.AddServiceVIPToName(vip, svcPort.Protocol, svc.Namespace, svc.Name)
 					ovn.handleExternalIPs(svc, svcPort, ips, targetPort)
 				}
 			}
@@ -95,7 +97,7 @@ func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 		targetPort := lbEps.Port
 		for _, svcPort := range svc.Spec.Ports {
 			if svcPort.Protocol == kapi.ProtocolUDP && svcPort.Name == svcPortName {
-				if util.ServiceTypeHasNodePort(svc) && config.Gateway.NodeportEnable {
+				if util.ServiceTypeHasNodePort(svc) {
 					err = ovn.createGatewaysVIP(string(svcPort.Protocol), svcPort.NodePort, targetPort, ips)
 					if err != nil {
 						logrus.Errorf("Error in creating Node Port for svc %s, node port: %d - %v\n", svc.Name, svcPort.NodePort, err)
@@ -116,6 +118,8 @@ func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 						logrus.Errorf("Error in creating Cluster IP for svc %s, target port: %d - %v\n", svc.Name, targetPort, err)
 						continue
 					}
+					vip := util.JoinHostPortInt32(svc.Spec.ClusterIP, svcPort.Port)
+					ovn.AddServiceVIPToName(vip, svcPort.Protocol, svc.Namespace, svc.Name)
 					ovn.handleExternalIPs(svc, svcPort, ips, targetPort)
 				}
 			}
@@ -124,34 +128,30 @@ func (ovn *Controller) AddEndpoints(ep *kapi.Endpoints) error {
 	return nil
 }
 
-func (ovn *Controller) handleNodePortLB(node *kapi.Node) bool {
+func (ovn *Controller) handleNodePortLB(node *kapi.Node) error {
 	physicalGateway := "GR_" + node.Name
 	var k8sNSLbTCP, k8sNSLbUDP, physicalIP string
 	if k8sNSLbTCP, _ = ovn.getGatewayLoadBalancer(physicalGateway, TCP); k8sNSLbTCP == "" {
-		logrus.Debugf("TCP load balancer for node %q does not yet exist", node.Name)
-		return false
+		return fmt.Errorf("TCP load balancer for node %q does not yet exist", node.Name)
 	}
 	if k8sNSLbUDP, _ = ovn.getGatewayLoadBalancer(physicalGateway, UDP); k8sNSLbUDP == "" {
-		logrus.Debugf("UDP load balancer for node %q does not yet exist", node.Name)
-		return false
+		return fmt.Errorf("UDP load balancer for node %q does not yet exist", node.Name)
 	}
 	if physicalIP, _ = ovn.getGatewayPhysicalIP(physicalGateway); physicalIP == "" {
-		logrus.Debugf("gateway physical IP for node %q does not yet exist", node.Name)
-		return false
+		return fmt.Errorf("gateway physical IP for node %q does not yet exist", node.Name)
 	}
-	namespaces, err := ovn.kube.GetNamespaces()
+	namespaces, err := ovn.watchFactory.GetNamespaces()
 	if err != nil {
-		logrus.Errorf("failed to get k8s namespaces: %v", err)
-		return false
+		return fmt.Errorf("failed to get k8s namespaces: %v", err)
 	}
-	for _, ns := range namespaces.Items {
-		endpoints, err := ovn.kube.GetEndpoints(ns.Name)
+	for _, ns := range namespaces {
+		endpoints, err := ovn.watchFactory.GetEndpoints(ns.Name)
 		if err != nil {
 			logrus.Errorf("failed to get k8s endpoints: %v", err)
 			continue
 		}
-		for _, ep := range endpoints.Items {
-			svc, err := ovn.kube.GetService(ep.Namespace, ep.Name)
+		for _, ep := range endpoints {
+			svc, err := ovn.watchFactory.GetService(ep.Namespace, ep.Name)
 			if err != nil {
 				continue
 			}
@@ -160,7 +160,7 @@ func (ovn *Controller) handleNodePortLB(node *kapi.Node) bool {
 			}
 			tcpPortMap := make(map[string]lbEndpoints)
 			udpPortMap := make(map[string]lbEndpoints)
-			ovn.getLbEndpoints(&ep, tcpPortMap, udpPortMap)
+			ovn.getLbEndpoints(ep, tcpPortMap, udpPortMap)
 			for svcPortName, lbEps := range tcpPortMap {
 				ips := lbEps.IPs
 				targetPort := lbEps.Port
@@ -191,7 +191,52 @@ func (ovn *Controller) handleNodePortLB(node *kapi.Node) bool {
 			}
 		}
 	}
-	return true
+	return nil
+}
+
+func (ovn *Controller) handleExternalIPsLB() {
+	namespaces, err := ovn.watchFactory.GetNamespaces()
+	if err != nil {
+		logrus.Errorf("failed to get k8s namespaces: %v", err)
+		return
+	}
+	for _, ns := range namespaces {
+		endpoints, err := ovn.watchFactory.GetEndpoints(ns.Name)
+		if err != nil {
+			logrus.Errorf("failed to get k8s endpoints: %v", err)
+			continue
+		}
+		for _, ep := range endpoints {
+			svc, err := ovn.watchFactory.GetService(ep.Namespace, ep.Name)
+			if err != nil {
+				continue
+			}
+			if len(svc.Spec.ExternalIPs) == 0 {
+				continue
+			}
+			tcpPortMap := make(map[string]lbEndpoints)
+			udpPortMap := make(map[string]lbEndpoints)
+			ovn.getLbEndpoints(ep, tcpPortMap, udpPortMap)
+			for svcPortName, lbEps := range tcpPortMap {
+				ips := lbEps.IPs
+				targetPort := lbEps.Port
+				for _, svcPort := range svc.Spec.Ports {
+					if svcPort.Protocol == kapi.ProtocolTCP && svcPort.Name == svcPortName {
+						ovn.handleExternalIPs(svc, svcPort, ips, targetPort)
+					}
+				}
+			}
+			for svcPortName, lbEps := range udpPortMap {
+				ips := lbEps.IPs
+				targetPort := lbEps.Port
+				for _, svcPort := range svc.Spec.Ports {
+					if svcPort.Protocol == kapi.ProtocolUDP && svcPort.Name == svcPortName {
+						ovn.handleExternalIPs(svc, svcPort, ips, targetPort)
+					}
+				}
+			}
+		}
+	}
 }
 
 func (ovn *Controller) handleExternalIPs(svc *kapi.Service, svcPort kapi.ServicePort, ips []string, targetPort int32) {
@@ -212,8 +257,15 @@ func (ovn *Controller) handleExternalIPs(svc *kapi.Service, svcPort kapi.Service
 	}
 }
 
+// keepEmptyLB returns true if empty load-balancer events is enabled and if the
+// service was idled.
+func keepEmptyLB(service *kapi.Service) bool {
+	_, ok := service.Annotations[OvnServiceIdledAt]
+	return config.Kubernetes.OVNEmptyLbEvents && ok
+}
+
 func (ovn *Controller) deleteEndpoints(ep *kapi.Endpoints) error {
-	svc, err := ovn.kube.GetService(ep.Namespace, ep.Name)
+	svc, err := ovn.watchFactory.GetService(ep.Namespace, ep.Name)
 	if err != nil {
 		// This is not necessarily an error. For e.g when a service is deleted,
 		// you will get endpoint delete event and the call to fetch service
@@ -233,13 +285,23 @@ func (ovn *Controller) deleteEndpoints(ep *kapi.Endpoints) error {
 				lb, err)
 			continue
 		}
-		key := fmt.Sprintf("\"%s:%d\"", svc.Spec.ClusterIP, svcPort.Port)
-		_, stderr, err := util.RunOVNNbctl("remove", "load_balancer", lb,
-			"vips", key)
-		if err != nil {
-			logrus.Errorf("Error in deleting endpoints for lb %s, "+
-				"stderr: %q (%v)", lb, stderr, err)
+
+		quotedHostPort := "\"" + util.JoinHostPortInt32(svc.Spec.ClusterIP, svcPort.Port) + "\""
+		if keepEmptyLB(svc) {
+			key := "vips:" + quotedHostPort + "=\"\""
+			_, stderr, err := util.RunOVNNbctl("set", "load_balancer", lb, key)
+			if err != nil {
+				logrus.Errorf("Error in deleting endpoints for lb %s, "+
+					"stderr: %q (%v)", lb, stderr, err)
+			}
+		} else {
+			_, stderr, err := util.RunOVNNbctl("remove", "load_balancer", lb,
+				"vips", quotedHostPort)
+			if err != nil {
+				logrus.Errorf("Error in deleting lb %s, stderr: %q (%v)", lb, stderr, err)
+			}
 		}
+
 	}
 	return nil
 }
