@@ -2,14 +2,13 @@ package ovn
 
 import (
 	"fmt"
-	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/factory"
+	"strings"
+
 	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/util"
 	"github.com/sirupsen/logrus"
 	kapi "k8s.io/api/core/v1"
 	knet "k8s.io/api/networking/v1"
-	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/client-go/tools/cache"
-	"strings"
 )
 
 func (oc *Controller) syncNetworkPoliciesPortGroup(
@@ -33,10 +32,10 @@ func (oc *Controller) syncNetworkPoliciesPortGroup(
 			// policy doesn't exist on k8s. Delete the port group
 			portGroupName := fmt.Sprintf("%s_%s", namespaceName, policyName)
 			hashedLocalPortGroup := hashedPortGroup(portGroupName)
-			oc.deletePortGroup(hashedLocalPortGroup)
+			deletePortGroup(hashedLocalPortGroup)
 
 			// delete the address sets for this policy from OVN
-			oc.deleteAddressSet(hashedAddressSet(addrSetName))
+			deleteAddressSet(hashedAddressSet(addrSetName))
 		}
 	})
 	if err != nil {
@@ -44,9 +43,7 @@ func (oc *Controller) syncNetworkPoliciesPortGroup(
 	}
 }
 
-func (oc *Controller) addACLAllow(np *namespacePolicy,
-	match, l4Match string, ipBlockCidr bool, gressNum int,
-	policyType knet.PolicyType) {
+func addACLAllow(np *namespacePolicy, match, l4Match string, ipBlockCidr bool, gressNum int, policyType knet.PolicyType) {
 	var direction, action string
 	direction = toLport
 	if policyType == knet.PolicyTypeIngress {
@@ -93,9 +90,7 @@ func (oc *Controller) addACLAllow(np *namespacePolicy,
 	}
 }
 
-func (oc *Controller) modifyACLAllow(namespace, policy,
-	oldMatch string, newMatch string, gressNum int,
-	policyType knet.PolicyType) {
+func modifyACLAllow(namespace, policy, oldMatch string, newMatch string, gressNum int, policyType knet.PolicyType) {
 	uuid, stderr, err := util.RunOVNNbctl("--data=bare", "--no-heading",
 		"--columns=_uuid", "find", "ACL", oldMatch,
 		fmt.Sprintf("external-ids:namespace=%s", namespace),
@@ -112,7 +107,7 @@ func (oc *Controller) modifyACLAllow(namespace, policy,
 	if uuid != "" {
 		// We already have an ACL. We will update it.
 		_, stderr, err = util.RunOVNNbctl("set", "acl", uuid,
-			fmt.Sprintf("%s", newMatch))
+			newMatch)
 		if err != nil {
 			logrus.Errorf("failed to modify the allow-from rule for "+
 				"namespace=%s, policy=%s, stderr: %q (%v)",
@@ -122,17 +117,16 @@ func (oc *Controller) modifyACLAllow(namespace, policy,
 	}
 }
 
-func (oc *Controller) addIPBlockACLDeny(np *namespacePolicy,
-	except, priority string, gressNum int, policyType knet.PolicyType) {
+func addIPBlockACLDeny(np *namespacePolicy, except, priority string, gressNum int, policyType knet.PolicyType) {
 	var match, l3Match, direction, lportMatch string
 	direction = toLport
 	if policyType == knet.PolicyTypeIngress {
 		lportMatch = fmt.Sprintf("outport == @%s", np.portGroupName)
-		l3Match = fmt.Sprintf("ip4.src == %s", except)
+		l3Match = fmt.Sprintf("%s.src == %s", ipMatch(), except)
 		match = fmt.Sprintf("match=\"%s && %s\"", lportMatch, l3Match)
 	} else {
 		lportMatch = fmt.Sprintf("inport == @%s", np.portGroupName)
-		l3Match = fmt.Sprintf("ip4.dst == %s", except)
+		l3Match = fmt.Sprintf("%s.dst == %s", ipMatch(), except)
 		match = fmt.Sprintf("match=\"%s && %s\"", lportMatch, l3Match)
 	}
 
@@ -166,21 +160,27 @@ func (oc *Controller) addIPBlockACLDeny(np *namespacePolicy,
 		logrus.Errorf("error executing create ACL command, stderr: %q, %+v",
 			stderr, err)
 	}
-	return
 }
 
-func (oc *Controller) addACLDenyPortGroup(portGroupUUID, portGroupName,
-	priority string, policyType knet.PolicyType) error {
-	var match, direction string
-	direction = toLport
+func getACLMatch(portGroupName, match string, policyType knet.PolicyType) string {
+	var aclMatch string
 	if policyType == knet.PolicyTypeIngress {
-		match = fmt.Sprintf("match=\"outport == @%s\"", portGroupName)
+		aclMatch = "outport == @" + portGroupName
 	} else {
-		match = fmt.Sprintf("match=\"inport == @%s\"", portGroupName)
+		aclMatch = "inport == @" + portGroupName
 	}
 
+	if match != "" {
+		aclMatch += " && " + match
+	}
+
+	return "match=\"" + aclMatch + "\""
+}
+
+func addACLPortGroup(portGroupUUID, portGroupName, direction, priority, match, action string, policyType knet.PolicyType) error {
+	match = getACLMatch(portGroupName, match, policyType)
 	uuid, stderr, err := util.RunOVNNbctl("--data=bare", "--no-heading",
-		"--columns=_uuid", "find", "ACL", match, "action=drop",
+		"--columns=_uuid", "find", "ACL", match, "action="+action,
 		fmt.Sprintf("external-ids:default-deny-policy-type=%s", policyType))
 	if err != nil {
 		return fmt.Errorf("find failed to get the default deny rule for "+
@@ -193,7 +193,7 @@ func (oc *Controller) addACLDenyPortGroup(portGroupUUID, portGroupName,
 
 	_, stderr, err = util.RunOVNNbctl("--id=@acl", "create", "acl",
 		fmt.Sprintf("priority=%s", priority),
-		fmt.Sprintf("direction=%s", direction), match, "action=drop",
+		fmt.Sprintf("direction=%s", direction), match, "action="+action,
 		fmt.Sprintf("external-ids:default-deny-policy-type=%s", policyType),
 		"--", "add", "port_group", portGroupUUID,
 		"acls", "@acl")
@@ -204,37 +204,62 @@ func (oc *Controller) addACLDenyPortGroup(portGroupUUID, portGroupName,
 	return nil
 }
 
-func (oc *Controller) addToACLDeny(portGroup, logicalPort string) {
-	logicalPortUUID := oc.getLogicalPortUUID(logicalPort)
-	if logicalPortUUID == "" {
-		return
+func deleteACLPortGroup(portGroupName, direction, priority, match, action string, policyType knet.PolicyType) error {
+	match = getACLMatch(portGroupName, match, policyType)
+	uuid, stderr, err := util.RunOVNNbctl("--data=bare", "--no-heading",
+		"--columns=_uuid", "find", "ACL", match, "action="+action,
+		fmt.Sprintf("external-ids:default-deny-policy-type=%s", policyType))
+	if err != nil {
+		return fmt.Errorf("find failed to get the rule for "+
+			"policy type %s stderr: %q (%v)", policyType, stderr, err)
+	}
+
+	if uuid == "" {
+		return nil
+	}
+
+	_, stderr, err = util.RunOVNNbctl("remove", "port_group",
+		portGroupName, "acls", uuid)
+	if err != nil {
+		return fmt.Errorf("remove failed to delete the rule for "+
+			"port_group=%s, stderr: %q (%v)", portGroupName, stderr, err)
+	}
+
+	return nil
+}
+
+func (oc *Controller) addToPortGroup(portGroup, logicalPort string) error {
+	logicalPortUUID, err := oc.getLogicalPortUUID(logicalPort)
+	if err != nil {
+		return err
 	}
 
 	_, stderr, err := util.RunOVNNbctl("--if-exists", "remove",
 		"port_group", portGroup, "ports", logicalPortUUID, "--",
 		"add", "port_group", portGroup, "ports", logicalPortUUID)
 	if err != nil {
-		logrus.Errorf("Failed to add logicalPort %s to portGroup %s "+
+		return fmt.Errorf("failed to add logicalPort %s to portGroup %s "+
 			"stderr: %q (%v)", logicalPort, portGroup, stderr, err)
 	}
+	return nil
 }
 
-func (oc *Controller) deleteFromACLDeny(portGroup, logicalPort string) {
-	logicalPortUUID := oc.getLogicalPortUUID(logicalPort)
-	if logicalPortUUID == "" {
-		return
+func (oc *Controller) deleteFromPortGroup(portGroup, logicalPort string) error {
+	logicalPortUUID, err := oc.getLogicalPortUUID(logicalPort)
+	if err != nil {
+		return err
 	}
 
 	_, stderr, err := util.RunOVNNbctl("--if-exists", "remove",
 		"port_group", portGroup, "ports", logicalPortUUID)
 	if err != nil {
-		logrus.Errorf("Failed to delete logicalPort %s to portGroup %s "+
+		return fmt.Errorf("failed to delete logicalPort %s to portGroup %s "+
 			"stderr: %q (%v)", logicalPort, portGroup, stderr, err)
 	}
+	return nil
 }
 
-func (oc *Controller) localPodAddACL(np *namespacePolicy,
-	gress *gressPolicy) {
+func localPodAddACL(np *namespacePolicy, gress *gressPolicy) {
 	l3Match := gress.getL3MatchFromAddressSet()
 
 	var lportMatch, cidrMatch string
@@ -248,8 +273,7 @@ func (oc *Controller) localPodAddACL(np *namespacePolicy,
 	// add deny acl rule with priority ipBlockDenyPriority (1010).
 	if len(gress.ipBlockCidr) > 0 && len(gress.ipBlockExcept) > 0 {
 		except := fmt.Sprintf("{%s}", strings.Join(gress.ipBlockExcept, ", "))
-		oc.addIPBlockACLDeny(np, except, ipBlockDenyPriority, gress.idx,
-			gress.policyType)
+		addIPBlockACLDeny(np, except, ipBlockDenyPriority, gress.idx, gress.policyType)
 	}
 
 	if len(gress.portPolicies) == 0 {
@@ -260,11 +284,9 @@ func (oc *Controller) localPodAddACL(np *namespacePolicy,
 		if len(gress.ipBlockCidr) > 0 {
 			// Add ACL allow rule for IPBlock CIDR
 			cidrMatch = gress.getMatchFromIPBlock(lportMatch, l4Match)
-			oc.addACLAllow(np, cidrMatch, l4Match,
-				true, gress.idx, gress.policyType)
+			addACLAllow(np, cidrMatch, l4Match, true, gress.idx, gress.policyType)
 		}
-		oc.addACLAllow(np, match, l4Match,
-			false, gress.idx, gress.policyType)
+		addACLAllow(np, match, l4Match, false, gress.idx, gress.policyType)
 	}
 	for _, port := range gress.portPolicies {
 		l4Match, err := port.getL4Match()
@@ -276,38 +298,34 @@ func (oc *Controller) localPodAddACL(np *namespacePolicy,
 		if len(gress.ipBlockCidr) > 0 {
 			// Add ACL allow rule for IPBlock CIDR
 			cidrMatch = gress.getMatchFromIPBlock(lportMatch, l4Match)
-			oc.addACLAllow(np, cidrMatch, l4Match,
-				true, gress.idx, gress.policyType)
+			addACLAllow(np, cidrMatch, l4Match, true, gress.idx, gress.policyType)
 		}
-		oc.addACLAllow(np, match, l4Match,
-			false, gress.idx, gress.policyType)
+		addACLAllow(np, match, l4Match, false, gress.idx, gress.policyType)
 	}
 }
 
-func (oc *Controller) createDefaultDenyPortGroup(policyType knet.PolicyType) {
+func (oc *Controller) createDefaultDenyPortGroup(policyType knet.PolicyType) error {
 	var portGroupName string
 	if policyType == knet.PolicyTypeIngress {
 		if oc.portGroupIngressDeny != "" {
-			return
+			return nil
 		}
 		portGroupName = "ingressDefaultDeny"
 	} else if policyType == knet.PolicyTypeEgress {
 		if oc.portGroupEgressDeny != "" {
-			return
+			return nil
 		}
 		portGroupName = "egressDefaultDeny"
 	}
-	portGroupUUID, err := oc.createPortGroup(portGroupName, portGroupName)
+	portGroupUUID, err := createPortGroup(portGroupName, portGroupName)
 	if err != nil {
-		logrus.Errorf("Failed to create port_group for %s (%v)",
+		return fmt.Errorf("Failed to create port_group for %s (%v)",
 			portGroupName, err)
-		return
 	}
-	err = oc.addACLDenyPortGroup(portGroupUUID, portGroupName,
-		defaultDenyPriority, policyType)
+	err = addACLPortGroup(portGroupUUID, portGroupName, toLport,
+		defaultDenyPriority, "", "drop", policyType)
 	if err != nil {
-		logrus.Errorf("Failed to create default deny port group %v", err)
-		return
+		return fmt.Errorf("Failed to create default deny port group %v", err)
 	}
 
 	if policyType == knet.PolicyTypeIngress {
@@ -315,16 +333,160 @@ func (oc *Controller) createDefaultDenyPortGroup(policyType knet.PolicyType) {
 	} else if policyType == knet.PolicyTypeEgress {
 		oc.portGroupEgressDeny = portGroupUUID
 	}
+	return nil
+}
+
+// Creates the match string used for ACLs allowing incoming multicast into a
+// namespace, that is, from IPs that are in the namespace's address set.
+func getMulticastACLMatch(ns string) string {
+	nsAddressSet := hashedAddressSet(ns)
+	return "ip4.src == $" + nsAddressSet + " && ip4.mcast"
+}
+
+// Returns the multicast port group name and hash for namespace 'ns'.
+func getMulticastPortGroup(ns string) (string, string) {
+	portGroupName := "mcastPortGroup-" + ns
+	return portGroupName, hashedPortGroup(portGroupName)
+}
+
+// Creates a policy to allow multicast traffic within 'ns':
+// - a port group containing all logical ports associated with 'ns'
+// - one "from-lport" ACL allowing egress multicast traffic from the pods
+//   in 'ns'
+// - one "to-lport" ACL allowing ingress multicast traffic to pods in 'ns'.
+//   This matches only traffic originated by pods in 'ns' (based on the
+//   namespace address set).
+func (oc *Controller) createMulticastAllowPolicy(ns string) error {
+	portGroupName, portGroupHash := getMulticastPortGroup(ns)
+	portGroupUUID, err := createPortGroup(portGroupName, portGroupHash)
+	if err != nil {
+		return fmt.Errorf("Failed to create port_group for %s (%v)",
+			portGroupName, err)
+	}
+
+	err = addACLPortGroup(portGroupUUID, portGroupHash, fromLport,
+		defaultMcastAllowPriority, "ip4.mcast", "allow",
+		knet.PolicyTypeEgress)
+	if err != nil {
+		return fmt.Errorf("Failed to create allow egress multicast ACL for %s (%v)",
+			ns, err)
+	}
+
+	err = addACLPortGroup(portGroupUUID, portGroupHash, toLport,
+		defaultMcastAllowPriority, getMulticastACLMatch(ns), "allow",
+		knet.PolicyTypeIngress)
+	if err != nil {
+		return fmt.Errorf("Failed to create allow ingress multicast ACL for %s (%v)",
+			ns, err)
+	}
+
+	// Add all ports from this namespace to the multicast allow group.
+	for _, portName := range oc.namespaceAddressSet[ns] {
+		if err := oc.podAddAllowMulticastPolicy(ns, portName); err != nil {
+			logrus.Warningf("failed to add port %s to port group ACL: %v", portName, err)
+		}
+	}
+
+	return nil
+}
+
+// Delete the policy to allow multicast traffic within 'ns'.
+func deleteMulticastAllowPolicy(ns string) error {
+	_, portGroupHash := getMulticastPortGroup(ns)
+
+	err := deleteACLPortGroup(portGroupHash, fromLport,
+		defaultMcastAllowPriority, "ip4.mcast", "allow",
+		knet.PolicyTypeEgress)
+	if err != nil {
+		return fmt.Errorf("Failed to delete allow egress multicast ACL for %s (%v)",
+			ns, err)
+	}
+
+	err = deleteACLPortGroup(portGroupHash, toLport,
+		defaultMcastAllowPriority, getMulticastACLMatch(ns), "allow",
+		knet.PolicyTypeIngress)
+	if err != nil {
+		return fmt.Errorf("Failed to delete allow ingress multicast ACL for %s (%v)",
+			ns, err)
+	}
+
+	deletePortGroup(portGroupHash)
+	return nil
+}
+
+// Creates a global default deny multicast policy:
+// - one ACL dropping egress multicast traffic from all pods: this is to
+//   protect OVN controller from processing IP multicast reports from nodes
+//   that are not allowed to receive multicast raffic.
+// - one ACL dropping ingress multicast traffic to all pods.
+func createDefaultDenyMulticastPolicy() error {
+	portGroupName := "mcastPortGroupDeny"
+	portGroupUUID, err := createPortGroup(portGroupName, portGroupName)
+	if err != nil {
+		return fmt.Errorf("Failed to create port_group for %s (%v)",
+			portGroupName, err)
+	}
+
+	// By default deny any egress multicast traffic from any pod. This drops
+	// IP multicast membership reports therefore denying any multicast traffic
+	// to be forwarded to pods.
+	err = addACLPortGroup(portGroupUUID, portGroupName, fromLport,
+		defaultMcastDenyPriority, "ip4.mcast", "drop", knet.PolicyTypeEgress)
+	if err != nil {
+		return fmt.Errorf("Failed to create default deny multicast egress ACL (%v)",
+			err)
+	}
+
+	// By default deny any ingress multicast traffic to any pod.
+	err = addACLPortGroup(portGroupUUID, portGroupName, toLport,
+		defaultMcastDenyPriority, "ip4.mcast", "drop", knet.PolicyTypeIngress)
+	if err != nil {
+		return fmt.Errorf("Failed to create default deny multicast ingress ACL (%v)",
+			err)
+	}
+
+	return nil
+}
+
+func (oc *Controller) podAddDefaultDenyMulticastPolicy(logicalPort string) error {
+	if err := oc.addToPortGroup("mcastPortGroupDeny", logicalPort); err != nil {
+		return fmt.Errorf("failed to add port %s to default multicast deny ACL: %v", logicalPort, err)
+	}
+	return nil
+}
+
+func (oc *Controller) podDeleteDefaultDenyMulticastPolicy(logicalPort string) error {
+	if err := oc.deleteFromPortGroup("mcastPortGroupDeny", logicalPort); err != nil {
+		return fmt.Errorf("failed to delete port %s from default multicast deny ACL: %v", logicalPort, err)
+	}
+	return nil
+}
+
+func (oc *Controller) podAddAllowMulticastPolicy(ns, logicalPort string) error {
+	_, portGroupHash := getMulticastPortGroup(ns)
+	return oc.addToPortGroup(portGroupHash, logicalPort)
+}
+
+func (oc *Controller) podDeleteAllowMulticastPolicy(ns, logicalPort string) error {
+	_, portGroupHash := getMulticastPortGroup(ns)
+	return oc.deleteFromPortGroup(portGroupHash, logicalPort)
 }
 
 func (oc *Controller) localPodAddDefaultDeny(
 	policy *knet.NetworkPolicy, logicalPort string) {
-
 	oc.lspMutex.Lock()
 	defer oc.lspMutex.Unlock()
 
-	oc.createDefaultDenyPortGroup(knet.PolicyTypeIngress)
-	oc.createDefaultDenyPortGroup(knet.PolicyTypeEgress)
+	err := oc.createDefaultDenyPortGroup(knet.PolicyTypeIngress)
+	if err != nil {
+		logrus.Errorf(err.Error())
+		return
+	}
+	err = oc.createDefaultDenyPortGroup(knet.PolicyTypeEgress)
+	if err != nil {
+		logrus.Errorf(err.Error())
+		return
+	}
 
 	// Default deny rule.
 	// 1. Any pod that matches a network policy should get a default
@@ -340,7 +502,9 @@ func (oc *Controller) localPodAddDefaultDeny(
 	// Handle condition 1 above.
 	if !(len(policy.Spec.PolicyTypes) == 1 && policy.Spec.PolicyTypes[0] == knet.PolicyTypeEgress) {
 		if oc.lspIngressDenyCache[logicalPort] == 0 {
-			oc.addToACLDeny(oc.portGroupIngressDeny, logicalPort)
+			if err := oc.addToPortGroup(oc.portGroupIngressDeny, logicalPort); err != nil {
+				logrus.Warningf("failed to add port %s to ingress deny ACL: %v", logicalPort, err)
+			}
 		}
 		oc.lspIngressDenyCache[logicalPort]++
 	}
@@ -349,7 +513,9 @@ func (oc *Controller) localPodAddDefaultDeny(
 	if (len(policy.Spec.PolicyTypes) == 1 && policy.Spec.PolicyTypes[0] == knet.PolicyTypeEgress) ||
 		len(policy.Spec.Egress) > 0 || len(policy.Spec.PolicyTypes) == 2 {
 		if oc.lspEgressDenyCache[logicalPort] == 0 {
-			oc.addToACLDeny(oc.portGroupEgressDeny, logicalPort)
+			if err := oc.addToPortGroup(oc.portGroupEgressDeny, logicalPort); err != nil {
+				logrus.Warningf("failed to add port %s to egress deny ACL: %v", logicalPort, err)
+			}
 		}
 		oc.lspEgressDenyCache[logicalPort]++
 	}
@@ -364,7 +530,9 @@ func (oc *Controller) localPodDelDefaultDeny(
 		if oc.lspIngressDenyCache[logicalPort] > 0 {
 			oc.lspIngressDenyCache[logicalPort]--
 			if oc.lspIngressDenyCache[logicalPort] == 0 {
-				oc.deleteFromACLDeny(oc.portGroupIngressDeny, logicalPort)
+				if err := oc.deleteFromPortGroup(oc.portGroupIngressDeny, logicalPort); err != nil {
+					logrus.Warningf("failed to remove port %s from ingress deny ACL: %v", logicalPort, err)
+				}
 			}
 		}
 	}
@@ -374,7 +542,9 @@ func (oc *Controller) localPodDelDefaultDeny(
 		if oc.lspEgressDenyCache[logicalPort] > 0 {
 			oc.lspEgressDenyCache[logicalPort]--
 			if oc.lspEgressDenyCache[logicalPort] == 0 {
-				oc.deleteFromACLDeny(oc.portGroupEgressDeny, logicalPort)
+				if err := oc.deleteFromPortGroup(oc.portGroupEgressDeny, logicalPort); err != nil {
+					logrus.Warningf("failed to remove port %s from egress deny ACL: %v", logicalPort, err)
+				}
 			}
 		}
 	}
@@ -385,20 +555,19 @@ func (oc *Controller) handleLocalPodSelectorAddFunc(
 	obj interface{}) {
 	pod := obj.(*kapi.Pod)
 
-	ipAddress := oc.getIPFromOvnAnnotation(pod.Annotations["ovn"])
-	if ipAddress == "" {
+	if _, err := util.UnmarshalPodAnnotation(pod.Annotations); err != nil {
 		return
 	}
 
-	logicalSwitch := pod.Spec.NodeName
-	if logicalSwitch == "" {
+	if pod.Spec.NodeName == "" {
 		return
 	}
 
 	// Get the logical port name.
-	logicalPort := fmt.Sprintf("%s_%s", pod.Namespace, pod.Name)
-	logicalPortUUID := oc.getLogicalPortUUID(logicalPort)
-	if logicalPortUUID == "" {
+	logicalPort := podLogicalPortName(pod)
+	logicalPortUUID, err := oc.getLogicalPortUUID(logicalPort)
+	if err != nil {
+		logrus.Errorf(err.Error())
 		return
 	}
 
@@ -435,14 +604,17 @@ func (oc *Controller) handleLocalPodSelectorDelFunc(
 	obj interface{}) {
 	pod := obj.(*kapi.Pod)
 
-	logicalSwitch := pod.Spec.NodeName
-	if logicalSwitch == "" {
+	if pod.Spec.NodeName == "" {
 		return
 	}
 
 	// Get the logical port name.
-	logicalPort := fmt.Sprintf("%s_%s", pod.Namespace, pod.Name)
-	logicalPortUUID := oc.getLogicalPortUUID(logicalPort)
+	logicalPort := podLogicalPortName(pod)
+	logicalPortUUID, err := oc.getLogicalPortUUID(logicalPort)
+	if err != nil {
+		logrus.Errorf(err.Error())
+		return
+	}
 
 	np.Lock()
 	defer np.Unlock()
@@ -457,6 +629,11 @@ func (oc *Controller) handleLocalPodSelectorDelFunc(
 	delete(np.localPods, logicalPort)
 	oc.localPodDelDefaultDeny(policy, logicalPort)
 
+	oc.lspMutex.Lock()
+	delete(oc.lspIngressDenyCache, logicalPort)
+	delete(oc.lspEgressDenyCache, logicalPort)
+	oc.lspMutex.Unlock()
+
 	if logicalPortUUID == "" || np.portGroupUUID == "" {
 		return
 	}
@@ -494,153 +671,6 @@ func (oc *Controller) handleLocalPodSelector(
 	np.podHandlerList = append(np.podHandlerList, h)
 }
 
-func (oc *Controller) handlePeerNamespaceAndPodSelector(
-	policy *knet.NetworkPolicy,
-	namespaceSelector *metav1.LabelSelector,
-	podSelector *metav1.LabelSelector,
-	addressSet string,
-	addressMap map[string]bool,
-	gress *gressPolicy,
-	np *namespacePolicy) {
-
-	namespaceHandler, err := oc.watchFactory.AddFilteredNamespaceHandler("",
-		namespaceSelector,
-		cache.ResourceEventHandlerFuncs{
-			AddFunc: func(obj interface{}) {
-				namespace := obj.(*kapi.Namespace)
-				np.Lock()
-				alreadyDeleted := np.deleted
-				np.Unlock()
-				if alreadyDeleted {
-					return
-				}
-
-				// The AddFilteredPodHandler call might call handlePeerPodSelectorAddUpdate
-				// on existing pods so we can't be holding the lock at this point
-				podHandler, err := oc.watchFactory.AddFilteredPodHandler(namespace.Name,
-					podSelector,
-					cache.ResourceEventHandlerFuncs{
-						AddFunc: func(obj interface{}) {
-							oc.handlePeerPodSelectorAddUpdate(policy, np, addressMap, addressSet, obj)
-						},
-						DeleteFunc: func(obj interface{}) {
-							oc.handlePeerPodSelectorDelete(policy, np, addressMap, addressSet, obj)
-						},
-						UpdateFunc: func(oldObj, newObj interface{}) {
-							oc.handlePeerPodSelectorAddUpdate(policy, np, addressMap, addressSet, newObj)
-						},
-					}, nil)
-				if err != nil {
-					logrus.Errorf("error watching pods in namespace %s for policy %s: %v", namespace.Name, policy.Name, err)
-					return
-				}
-				np.Lock()
-				defer np.Unlock()
-				if np.deleted {
-					_ = oc.watchFactory.RemovePodHandler(podHandler)
-					return
-				}
-				np.podHandlerList = append(np.podHandlerList, podHandler)
-			},
-			DeleteFunc: func(obj interface{}) {
-				return
-			},
-			UpdateFunc: func(oldObj, newObj interface{}) {
-				return
-			},
-		}, nil)
-	if err != nil {
-		logrus.Errorf("error watching namespaces for policy %s: %v",
-			policy.Name, err)
-		return
-	}
-	np.nsHandlerList = append(np.nsHandlerList, namespaceHandler)
-}
-
-func (oc *Controller) handlePeerPodSelectorAddUpdate(
-	policy *knet.NetworkPolicy, np *namespacePolicy,
-	addressMap map[string]bool, addressSet string,
-	obj interface{}) {
-
-	pod := obj.(*kapi.Pod)
-	ipAddress := oc.getIPFromOvnAnnotation(pod.Annotations["ovn"])
-	if ipAddress == "" || addressMap[ipAddress] {
-		return
-	}
-
-	np.Lock()
-	defer np.Unlock()
-	if np.deleted {
-		return
-	}
-
-	addressMap[ipAddress] = true
-	addresses := make([]string, 0, len(addressMap))
-	for k := range addressMap {
-		addresses = append(addresses, k)
-	}
-	oc.setAddressSet(addressSet, addresses)
-
-}
-
-func (oc *Controller) handlePeerPodSelectorDelete(
-	policy *knet.NetworkPolicy, np *namespacePolicy,
-	addressMap map[string]bool, addressSet string,
-	obj interface{}) {
-
-	pod := obj.(*kapi.Pod)
-
-	ipAddress := oc.getIPFromOvnAnnotation(pod.Annotations["ovn"])
-	if ipAddress == "" {
-		return
-	}
-
-	np.Lock()
-	defer np.Unlock()
-	if np.deleted {
-		return
-	}
-
-	if !addressMap[ipAddress] {
-		return
-	}
-
-	delete(addressMap, ipAddress)
-
-	addresses := make([]string, 0, len(addressMap))
-	for k := range addressMap {
-		addresses = append(addresses, k)
-	}
-	oc.setAddressSet(addressSet, addresses)
-}
-
-func (oc *Controller) handlePeerPodSelector(
-	policy *knet.NetworkPolicy, podSelector *metav1.LabelSelector,
-	addressSet string, addressMap map[string]bool, np *namespacePolicy) {
-
-	h, err := oc.watchFactory.AddFilteredPodHandler(policy.Namespace,
-		podSelector,
-		cache.ResourceEventHandlerFuncs{
-			AddFunc: func(obj interface{}) {
-				oc.handlePeerPodSelectorAddUpdate(policy, np, addressMap, addressSet, obj)
-			},
-			DeleteFunc: func(obj interface{}) {
-				oc.handlePeerPodSelectorDelete(policy, np, addressMap, addressSet, obj)
-			},
-			UpdateFunc: func(oldObj, newObj interface{}) {
-				oc.handlePeerPodSelectorAddUpdate(policy, np, addressMap, addressSet, newObj)
-			},
-		}, nil)
-	if err != nil {
-		logrus.Errorf("error watching peer pods for policy %s in namespace %s: %v",
-			policy.Name, policy.Namespace, err)
-		return
-	}
-
-	np.podHandlerList = append(np.podHandlerList, h)
-
-}
-
 func (oc *Controller) handlePeerNamespaceSelectorModify(
 	gress *gressPolicy, np *namespacePolicy, oldl3Match, newl3Match string) {
 
@@ -655,8 +685,7 @@ func (oc *Controller) handlePeerNamespaceSelectorModify(
 			lportMatch)
 		newMatch := fmt.Sprintf("match=\"%s && %s\"", newl3Match,
 			lportMatch)
-		oc.modifyACLAllow(np.namespace, np.name,
-			oldMatch, newMatch, gress.idx, gress.policyType)
+		modifyACLAllow(np.namespace, np.name, oldMatch, newMatch, gress.idx, gress.policyType)
 	}
 	for _, port := range gress.portPolicies {
 		l4Match, err := port.getL4Match()
@@ -667,62 +696,12 @@ func (oc *Controller) handlePeerNamespaceSelectorModify(
 			oldl3Match, l4Match, lportMatch)
 		newMatch := fmt.Sprintf("match=\"%s && %s && %s\"",
 			newl3Match, l4Match, lportMatch)
-		oc.modifyACLAllow(np.namespace, np.name,
-			oldMatch, newMatch, gress.idx, gress.policyType)
+		modifyACLAllow(np.namespace, np.name, oldMatch, newMatch, gress.idx, gress.policyType)
 	}
 }
 
-func (oc *Controller) handlePeerNamespaceSelector(
-	policy *knet.NetworkPolicy,
-	namespaceSelector *metav1.LabelSelector,
-	gress *gressPolicy, np *namespacePolicy) {
-
-	h, err := oc.watchFactory.AddFilteredNamespaceHandler("",
-		namespaceSelector,
-		cache.ResourceEventHandlerFuncs{
-			AddFunc: func(obj interface{}) {
-				namespace := obj.(*kapi.Namespace)
-				np.Lock()
-				defer np.Unlock()
-				if np.deleted {
-					return
-				}
-				hashedAddressSet := hashedAddressSet(namespace.Name)
-				oldL3Match, newL3Match, added := gress.addAddressSet(hashedAddressSet)
-				if added {
-					oc.handlePeerNamespaceSelectorModify(gress,
-						np, oldL3Match, newL3Match)
-				}
-			},
-			DeleteFunc: func(obj interface{}) {
-				namespace := obj.(*kapi.Namespace)
-				np.Lock()
-				defer np.Unlock()
-				if np.deleted {
-					return
-				}
-				hashedAddressSet := hashedAddressSet(namespace.Name)
-				oldL3Match, newL3Match, removed := gress.delAddressSet(hashedAddressSet)
-				if removed {
-					oc.handlePeerNamespaceSelectorModify(gress,
-						np, oldL3Match, newL3Match)
-				}
-			},
-			UpdateFunc: func(oldObj, newObj interface{}) {
-				return
-			},
-		}, nil)
-	if err != nil {
-		logrus.Errorf("error watching namespaces for policy %s: %v",
-			policy.Name, err)
-		return
-	}
-
-	np.nsHandlerList = append(np.nsHandlerList, h)
-
-}
-
-// AddNetworkPolicy adds network policy and create corresponding acl rules
+// addNetworkPolicyPortGroup creates and applies OVN ACLs to pod logical switch
+// ports from Kubernetes NetworkPolicy objects using OVN Port Groups
 func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 	logrus.Infof("Adding network policy %s in namespace %s", policy.Name,
 		policy.Namespace)
@@ -739,22 +718,14 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 		return
 	}
 
-	np := &namespacePolicy{}
-	np.name = policy.Name
-	np.namespace = policy.Namespace
-	np.ingressPolicies = make([]*gressPolicy, 0)
-	np.egressPolicies = make([]*gressPolicy, 0)
-	np.podHandlerList = make([]*factory.Handler, 0)
-	np.nsHandlerList = make([]*factory.Handler, 0)
-	np.localPods = make(map[string]bool)
+	np := NewNamespacePolicy(policy)
 
 	// Create a port group for the policy. All the pods that this policy
 	// selects will be eventually added to this port group.
 	readableGroupName := fmt.Sprintf("%s_%s", policy.Namespace, policy.Name)
 	np.portGroupName = hashedPortGroup(readableGroupName)
 
-	np.portGroupUUID, err = oc.createPortGroup(readableGroupName,
-		np.portGroupName)
+	np.portGroupUUID, err = createPortGroup(readableGroupName, np.portGroupName)
 	if err != nil {
 		logrus.Errorf("Failed to create port_group for network policy %s in "+
 			"namespace %s", policy.Name, policy.Namespace)
@@ -784,7 +755,7 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 				policy.Name, "ingress", i)
 
 			hashedLocalAddressSet = hashedAddressSet(localPeerPods)
-			oc.createAddressSet(localPeerPods, hashedLocalAddressSet, nil)
+			createAddressSet(localPeerPods, hashedLocalAddressSet, nil)
 			ingress.addAddressSet(hashedLocalAddressSet)
 		}
 
@@ -795,22 +766,23 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 			}
 		}
 
-		oc.localPodAddACL(np, ingress)
+		localPodAddACL(np, ingress)
 
 		for _, fromJSON := range ingressJSON.From {
 			if fromJSON.NamespaceSelector != nil && fromJSON.PodSelector != nil {
 				// For each rule that contains both peer namespace selector and
 				// peer pod selector, we create a watcher for each matching namespace
 				// that populates the addressSet
-				oc.handlePeerNamespaceAndPodSelector(policy,
+				oc.handlePeerNamespaceAndPodSelector(policy, ingress,
 					fromJSON.NamespaceSelector, fromJSON.PodSelector,
-					hashedLocalAddressSet, peerPodAddressMap, ingress, np)
+					hashedLocalAddressSet, peerPodAddressMap, np)
 
 			} else if fromJSON.NamespaceSelector != nil {
 				// For each peer namespace selector, we create a watcher that
 				// populates ingress.peerAddressSets
 				oc.handlePeerNamespaceSelector(policy,
-					fromJSON.NamespaceSelector, ingress, np)
+					fromJSON.NamespaceSelector, ingress, np,
+					oc.handlePeerNamespaceSelectorModify)
 			} else if fromJSON.PodSelector != nil {
 				// For each peer pod selector, we create a watcher that
 				// populates the addressSet
@@ -844,7 +816,7 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 				policy.Name, "egress", i)
 
 			hashedLocalAddressSet = hashedAddressSet(localPeerPods)
-			oc.createAddressSet(localPeerPods, hashedLocalAddressSet, nil)
+			createAddressSet(localPeerPods, hashedLocalAddressSet, nil)
 			egress.addAddressSet(hashedLocalAddressSet)
 		}
 
@@ -855,22 +827,23 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 			}
 		}
 
-		oc.localPodAddACL(np, egress)
+		localPodAddACL(np, egress)
 
 		for _, toJSON := range egressJSON.To {
 			if toJSON.NamespaceSelector != nil && toJSON.PodSelector != nil {
 				// For each rule that contains both peer namespace selector and
 				// peer pod selector, we create a watcher for each matching namespace
 				// that populates the addressSet
-				oc.handlePeerNamespaceAndPodSelector(policy,
+				oc.handlePeerNamespaceAndPodSelector(policy, egress,
 					toJSON.NamespaceSelector, toJSON.PodSelector,
-					hashedLocalAddressSet, peerPodAddressMap, egress, np)
+					hashedLocalAddressSet, peerPodAddressMap, np)
 
 			} else if toJSON.NamespaceSelector != nil {
 				// For each peer namespace selector, we create a watcher that
 				// populates egress.peerAddressSets
-				go oc.handlePeerNamespaceSelector(policy,
-					toJSON.NamespaceSelector, egress, np)
+				oc.handlePeerNamespaceSelector(policy,
+					toJSON.NamespaceSelector, egress, np,
+					oc.handlePeerNamespaceSelectorModify)
 			} else if toJSON.PodSelector != nil {
 				// For each peer pod selector, we create a watcher that
 				// populates the addressSet
@@ -886,8 +859,6 @@ func (oc *Controller) addNetworkPolicyPortGroup(policy *knet.NetworkPolicy) {
 	// For all the pods in the local namespace that this policy
 	// effects, add them to the port group.
 	oc.handleLocalPodSelector(policy, np)
-
-	return
 }
 
 func (oc *Controller) deleteNetworkPolicyPortGroup(
@@ -910,23 +881,6 @@ func (oc *Controller) deleteNetworkPolicyPortGroup(
 	// Mark the policy as deleted.
 	np.deleted = true
 
-	// Go through each ingress rule.  For each ingress rule, delete the
-	// addressSet for the local peer pods.
-	for i := range np.ingressPolicies {
-		localPeerPods := fmt.Sprintf("%s.%s.%s.%d", policy.Namespace,
-			policy.Name, "ingress", i)
-		hashedAddressSet := hashedAddressSet(localPeerPods)
-		oc.deleteAddressSet(hashedAddressSet)
-	}
-	// Go through each egress rule.  For each egress rule, delete the
-	// addressSet for the local peer pods.
-	for i := range np.egressPolicies {
-		localPeerPods := fmt.Sprintf("%s.%s.%s.%d", policy.Namespace,
-			policy.Name, "egress", i)
-		hashedAddressSet := hashedAddressSet(localPeerPods)
-		oc.deleteAddressSet(hashedAddressSet)
-	}
-
 	// We should now stop all the handlers go routines.
 	oc.shutdownHandlers(np)
 
@@ -935,9 +889,24 @@ func (oc *Controller) deleteNetworkPolicyPortGroup(
 	}
 
 	// Delete the port group
-	oc.deletePortGroup(np.portGroupName)
+	deletePortGroup(np.portGroupName)
+
+	// Go through each ingress rule.  For each ingress rule, delete the
+	// addressSet for the local peer pods.
+	for i := range np.ingressPolicies {
+		localPeerPods := fmt.Sprintf("%s.%s.%s.%d", policy.Namespace,
+			policy.Name, "ingress", i)
+		hashedAddressSet := hashedAddressSet(localPeerPods)
+		deleteAddressSet(hashedAddressSet)
+	}
+	// Go through each egress rule.  For each egress rule, delete the
+	// addressSet for the local peer pods.
+	for i := range np.egressPolicies {
+		localPeerPods := fmt.Sprintf("%s.%s.%s.%d", policy.Namespace,
+			policy.Name, "egress", i)
+		hashedAddressSet := hashedAddressSet(localPeerPods)
+		deleteAddressSet(hashedAddressSet)
+	}
 
 	oc.namespacePolicies[policy.Namespace][policy.Name] = nil
-
-	return
 }
