@@ -2,25 +2,21 @@ package util
 
 import (
 	"fmt"
+	"net"
 	"strings"
 
 	"github.com/sirupsen/logrus"
-
-	"github.com/ovn-org/ovn-kubernetes/go-controller/pkg/config"
 )
 
 // GatewayCleanup removes all the NB DB objects created for a node's gateway
-func GatewayCleanup(nodeName string) error {
+func GatewayCleanup(nodeName string, nodeSubnet *net.IPNet) error {
 	// Get the cluster router
-	clusterRouter, err := GetK8sClusterRouter()
-	if err != nil {
-		return fmt.Errorf("failed to get cluster router")
-	}
-
+	clusterRouter := GetK8sClusterRouter()
 	gatewayRouter := fmt.Sprintf("GR_%s", nodeName)
 
 	// Get the gateway router port's IP address (connected to join switch)
 	var routerIP string
+	var nextHops []string
 	routerIPNetwork, stderr, err := RunOVNNbctl("--if-exist", "get",
 		"logical_router_port", "rtoj-"+gatewayRouter, "networks")
 	if err != nil {
@@ -34,42 +30,26 @@ func GatewayCleanup(nodeName string) error {
 			routerIP = strings.Split(routerIPNetwork, "/")[0]
 		}
 	}
-
 	if routerIP != "" {
-		// Get a list of all the routes in cluster router with this gateway
-		// Router as the next hop.
-		var uuids string
-		uuids, stderr, err = RunOVNNbctl("--data=bare", "--no-heading",
-			"--columns=_uuid", "find", "logical_router_static_route",
-			"nexthop="+routerIP)
-		if err != nil {
-			return fmt.Errorf("Failed to fetch all routes with gateway "+
-				"router %s as nexthop, stderr: %q, "+
-				"error: %v", gatewayRouter, stderr, err)
-		}
-
-		// Remove all the routes in cluster router with this gateway Router
-		// as the nexthop.
-		routes := strings.Fields(uuids)
-		for _, route := range routes {
-			_, stderr, err = RunOVNNbctl("--if-exists", "remove",
-				"logical_router", clusterRouter, "static_routes", route)
-			if err != nil {
-				logrus.Errorf("Failed to delete static route %s"+
-					", stderr: %q, err = %v", route, stderr, err)
-				continue
-			}
-		}
+		nextHops = append(nextHops, routerIP)
 	}
 
-	// Remove the patch port that connects join switch to gateway router
-	_, stderr, err = RunOVNNbctl("--if-exist", "lsp-del", "jtor-"+gatewayRouter)
+	if nodeSubnet != nil {
+		_, mgtPortIP := GetNodeWellKnownAddresses(nodeSubnet)
+		if mgtPortIP.IP.String() != "" {
+			nextHops = append(nextHops, mgtPortIP.IP.String())
+		}
+	}
+	staticRouteCleanup(clusterRouter, nextHops)
+
+	// Remove the join switch that connects ovn_cluster_router to gateway router
+	_, stderr, err = RunOVNNbctl("--if-exist", "ls-del", "join_"+nodeName)
 	if err != nil {
-		return fmt.Errorf("Failed to delete logical switch port jtor-%s, "+
-			"stderr: %q, error: %v", gatewayRouter, stderr, err)
+		return fmt.Errorf("Failed to delete the join logical switch join_%s, "+
+			"stderr: %q, error: %v", nodeName, stderr, err)
 	}
 
-	// Remove any gateway routers associated with nodeName
+	// Remove the gateway router associated with nodeName
 	_, stderr, err = RunOVNNbctl("--if-exist", "lr-del",
 		gatewayRouter)
 	if err != nil {
@@ -86,17 +66,26 @@ func GatewayCleanup(nodeName string) error {
 			"error: %v", externalSwitch, stderr, err)
 	}
 
-	if config.Gateway.NodeportEnable {
-		//Remove the TCP, UDP load-balancers created for north-south traffic for gateway router.
-		k8sNSLbTCP, k8sNSLbUDP, err := getGatewayLoadBalancers(gatewayRouter)
-		if err != nil {
-			return err
-		}
+	// Remove the patch port on the distributed router that connects to join switch
+	_, stderr, err = RunOVNNbctl("--if-exist", "lrp-del", "dtoj-"+nodeName)
+	if err != nil {
+		return fmt.Errorf("Failed to delete the patch port dtoj-%s on distributed router "+
+			"stderr: %q, error: %v", nodeName, stderr, err)
+	}
+
+	// If exists, remove the TCP, UDP load-balancers created for north-south traffic for gateway router.
+	k8sNSLbTCP, k8sNSLbUDP, err := getGatewayLoadBalancers(gatewayRouter)
+	if err != nil {
+		return err
+	}
+	if k8sNSLbTCP != "" {
 		_, stderr, err = RunOVNNbctl("lb-del", k8sNSLbTCP)
 		if err != nil {
 			return fmt.Errorf("Failed to delete Gateway router TCP load balancer %s, stderr: %q, "+
 				"error: %v", k8sNSLbTCP, stderr, err)
 		}
+	}
+	if k8sNSLbUDP != "" {
 		_, stderr, err = RunOVNNbctl("lb-del", k8sNSLbUDP)
 		if err != nil {
 			return fmt.Errorf("Failed to delete Gateway router UDP load balancer %s, stderr: %q, "+
@@ -105,3 +94,31 @@ func GatewayCleanup(nodeName string) error {
 	}
 	return nil
 }
+
+func staticRouteCleanup(clusterRouter string, nextHops []string) {
+	for _, nextHop := range nextHops {
+		// Get a list of all the routes in cluster router with the next hop IP.
+		var uuids string
+		uuids, stderr, err := RunOVNNbctl("--data=bare", "--no-heading",
+			"--columns=_uuid", "find", "logical_router_static_route",
+			"nexthop=\""+nextHop+"\"")
+		if err != nil {
+			logrus.Errorf("Failed to fetch all routes with "+
+				"IP %s as nexthop, stderr: %q, "+
+				"error: %v", nextHop, stderr, err)
+			continue
+		}
+
+		// Remove all the routes in cluster router with this IP as the nexthop.
+		routes := strings.Fields(uuids)
+		for _, route := range routes {
+			_, stderr, err = RunOVNNbctl("--if-exists", "remove",
+				"logical_router", clusterRouter, "static_routes", route)
+			if err != nil {
+				logrus.Errorf("Failed to delete static route %s"+
+					", stderr: %q, err = %v", route, stderr, err)
+				continue
+			}
+		}
+	}
+}
