@@ -25,8 +25,10 @@ const (
 	ovsCommandTimeout = 15
 	ovsVsctlCommand   = "ovs-vsctl"
 	ovsOfctlCommand   = "ovs-ofctl"
+	ovsAppctlCommand  = "ovs-appctl"
 	ovnNbctlCommand   = "ovn-nbctl"
 	ovnSbctlCommand   = "ovn-sbctl"
+	ovnAppctlCommand  = "ovn-appctl"
 	ipCommand         = "ip"
 	powershellCommand = "powershell"
 	netshCommand      = "netsh"
@@ -37,6 +39,12 @@ const (
 	windowsOS         = "windows"
 )
 
+const (
+	nbdbCtlSock   = "ovnnb_db.ctl"
+	sbdbCtlSock   = "ovnsb_db.ctl"
+	northdCtlSock = "ovn-northd.ctl"
+)
+
 func runningPlatform() (string, error) {
 	if runtime.GOOS == windowsOS {
 		return windowsOS, nil
@@ -79,8 +87,12 @@ type execHelper struct {
 	exec           kexec.Interface
 	ofctlPath      string
 	vsctlPath      string
+	appctlPath     string
+	ovnappctlPath  string
 	nbctlPath      string
 	sbctlPath      string
+	ovnctlPath     string
+	ovnRunDir      string
 	ipPath         string
 	powershellPath string
 	netshPath      string
@@ -92,9 +104,11 @@ var runner *execHelper
 // SetExec validates executable paths and saves the given exec interface
 // to be used for running various OVS and OVN utilites
 func SetExec(exec kexec.Interface) error {
-	var err error
+	err := SetExecWithoutOVS(exec)
+	if err != nil {
+		return err
+	}
 
-	runner = &execHelper{exec: exec}
 	runner.ofctlPath, err = exec.LookPath(ovsOfctlCommand)
 	if err != nil {
 		return err
@@ -103,6 +117,26 @@ func SetExec(exec kexec.Interface) error {
 	if err != nil {
 		return err
 	}
+	runner.appctlPath, err = exec.LookPath(ovsAppctlCommand)
+	if err != nil {
+		return err
+	}
+
+	runner.ovnappctlPath, err = exec.LookPath(ovnAppctlCommand)
+	if err != nil {
+		// If ovn-appctl command is not available then fall back to
+		// ovs-appctl. It also means OVN is using the rundir of
+		// openvswitch.
+		runner.ovnappctlPath = runner.appctlPath
+		runner.ovnctlPath = "/usr/share/openvswitch/scripts/ovn-ctl"
+		runner.ovnRunDir = "/var/run/openvswitch/"
+	} else {
+		// If ovn-appctl command is available, it means OVN
+		// has its own separate rundir, logdir, sharedir.
+		runner.ovnctlPath = "/usr/share/ovn/scripts/ovn-ctl"
+		runner.ovnRunDir = "/var/run/ovn/"
+	}
+
 	runner.nbctlPath, err = exec.LookPath(ovnNbctlCommand)
 	if err != nil {
 		return err
@@ -111,6 +145,16 @@ func SetExec(exec kexec.Interface) error {
 	if err != nil {
 		return err
 	}
+
+	return nil
+}
+
+// SetExecWithoutOVS validates executable paths excluding OVS/OVN binaries and
+// saves the given exec interface to be used for running various utilites
+func SetExecWithoutOVS(exec kexec.Interface) error {
+	var err error
+
+	runner = &execHelper{exec: exec}
 	if runtime.GOOS == windowsOS {
 		runner.powershellPath, err = exec.LookPath(powershellCommand)
 		if err != nil {
@@ -133,6 +177,26 @@ func SetExec(exec kexec.Interface) error {
 	return nil
 }
 
+// SetSpecificExec validates executable paths for selected commands. It also saves the given
+// exec interface to be used for running selected commands
+func SetSpecificExec(exec kexec.Interface, commands ...string) error {
+	var err error
+
+	runner = &execHelper{exec: exec}
+	for _, command := range commands {
+		switch command {
+		case ovsVsctlCommand:
+			runner.vsctlPath, err = exec.LookPath(ovsVsctlCommand)
+			if err != nil {
+				return err
+			}
+		default:
+			return fmt.Errorf("unknown command: %q", command)
+		}
+	}
+	return nil
+}
+
 // GetExec returns the exec interface which can be used for running commands directly.
 // Only use for passing an exec interface into pkg/config which cannot call this
 // function directly because this module imports pkg/config already.
@@ -142,10 +206,10 @@ func GetExec() kexec.Interface {
 
 var runCounter uint64
 
-func run(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
+func runCmd(cmd kexec.Cmd, cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 	stdout := &bytes.Buffer{}
 	stderr := &bytes.Buffer{}
-	cmd := runner.exec.Command(cmdPath, args...)
+
 	cmd.SetStdout(stdout)
 	cmd.SetStderr(stderr)
 
@@ -162,6 +226,17 @@ func run(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 	return stdout, stderr, err
 }
 
+func run(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
+	cmd := runner.exec.Command(cmdPath, args...)
+	return runCmd(cmd, cmdPath, args...)
+}
+
+func runWithEnvVars(cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
+	cmd := runner.exec.Command(cmdPath, args...)
+	cmd.SetEnv(envVars)
+	return runCmd(cmd, cmdPath, args...)
+}
+
 // RunOVSOfctl runs a command via ovs-ofctl.
 func RunOVSOfctl(args ...string) (string, string, error) {
 	stdout, stderr, err := run(runner.ofctlPath, args...)
@@ -176,13 +251,21 @@ func RunOVSVsctl(args ...string) (string, string, error) {
 	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
 }
 
+// RunOVSAppctl runs a command via ovs-appctl.
+func RunOVSAppctl(args ...string) (string, string, error) {
+	cmdArgs := []string{fmt.Sprintf("--timeout=%d", ovsCommandTimeout)}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := run(runner.appctlPath, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
 // Run the ovn-ctl command and retry if "Connection refused"
 // poll waitng for service to become available
-func runOVNretry(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
+func runOVNretry(cmdPath string, envVars []string, args ...string) (*bytes.Buffer, *bytes.Buffer, error) {
 
 	retriesLeft := 200
 	for {
-		stdout, stderr, err := run(cmdPath, args...)
+		stdout, stderr, err := runWithEnvVars(cmdPath, envVars, args...)
 		if err == nil {
 			return stdout, stderr, err
 		}
@@ -202,36 +285,55 @@ func runOVNretry(cmdPath string, args ...string) (*bytes.Buffer, *bytes.Buffer,
 	}
 }
 
+func getNbctlArgsAndEnv(timeout int, args ...string) ([]string, []string) {
+	var cmdArgs, envVars []string
+
+	if config.NbctlDaemonMode {
+		// when ovn-nbctl is running in a "daemon mode", the user first starts
+		// ovn-nbctl running in the background and afterward uses the daemon to execute
+		// operations. The client needs to use the control socket and set the path to the
+		// control socket in environment variable OVN_NB_DAEMON
+		pid, err := ioutil.ReadFile(runner.ovnRunDir + "ovn-nbctl.pid")
+		if err == nil {
+			envVars = append(envVars,
+				fmt.Sprintf("OVN_NB_DAEMON=%sovn-nbctl.%s.ctl", runner.ovnRunDir,
+					strings.Trim(string(pid), " \n")))
+			logrus.Debugf("using ovn-nbctl daemon mode at %s", envVars)
+			cmdArgs = append(cmdArgs, fmt.Sprintf("--timeout=%d", timeout))
+			cmdArgs = append(cmdArgs, args...)
+			return cmdArgs, envVars
+		}
+		logrus.Warningf("failed to retrieve ovn-nbctl daemon's control socket " +
+			"so resorting to non-daemon mode")
+	}
+
+	if config.OvnNorth.Scheme == config.OvnDBSchemeSSL {
+		cmdArgs = append(cmdArgs,
+			fmt.Sprintf("--private-key=%s", config.OvnNorth.PrivKey),
+			fmt.Sprintf("--certificate=%s", config.OvnNorth.Cert),
+			fmt.Sprintf("--bootstrap-ca-cert=%s", config.OvnNorth.CACert),
+			fmt.Sprintf("--db=%s", config.OvnNorth.GetURL()))
+	} else if config.OvnNorth.Scheme == config.OvnDBSchemeTCP {
+		cmdArgs = append(cmdArgs, fmt.Sprintf("--db=%s", config.OvnNorth.GetURL()))
+	}
+	cmdArgs = append(cmdArgs, fmt.Sprintf("--timeout=%d", timeout))
+	cmdArgs = append(cmdArgs, args...)
+	return cmdArgs, envVars
+}
+
 // RunOVNNbctlUnix runs command via ovn-nbctl, with ovn-nbctl using the unix
 // domain sockets to connect to the ovsdb-server backing the OVN NB database.
 func RunOVNNbctlUnix(args ...string) (string, string, error) {
-	cmdArgs := []string{fmt.Sprintf("--timeout=%d", ovsCommandTimeout)}
-	cmdArgs = append(cmdArgs, args...)
-	stdout, stderr, err := runOVNretry(runner.nbctlPath, cmdArgs...)
+	cmdArgs, envVars := getNbctlArgsAndEnv(ovsCommandTimeout, args...)
+	stdout, stderr, err := runOVNretry(runner.nbctlPath, envVars, cmdArgs...)
 	return strings.Trim(strings.TrimFunc(stdout.String(), unicode.IsSpace), "\""),
 		stderr.String(), err
 }
 
 // RunOVNNbctlWithTimeout runs command via ovn-nbctl with a specific timeout
-func RunOVNNbctlWithTimeout(timeout int, args ...string) (string, string,
-	error) {
-	var cmdArgs []string
-	if config.OvnNorth.Scheme == config.OvnDBSchemeSSL {
-		cmdArgs = []string{
-			fmt.Sprintf("--private-key=%s", config.OvnNorth.PrivKey),
-			fmt.Sprintf("--certificate=%s", config.OvnNorth.Cert),
-			fmt.Sprintf("--bootstrap-ca-cert=%s", config.OvnNorth.CACert),
-			fmt.Sprintf("--db=%s", config.OvnNorth.GetURL()),
-		}
-	} else if config.OvnNorth.Scheme == config.OvnDBSchemeTCP {
-		cmdArgs = []string{
-			fmt.Sprintf("--db=%s", config.OvnNorth.GetURL()),
-		}
-	}
-
-	cmdArgs = append(cmdArgs, fmt.Sprintf("--timeout=%d", timeout))
-	cmdArgs = append(cmdArgs, args...)
-	stdout, stderr, err := runOVNretry(runner.nbctlPath, cmdArgs...)
+func RunOVNNbctlWithTimeout(timeout int, args ...string) (string, string, error) {
+	cmdArgs, envVars := getNbctlArgsAndEnv(timeout, args...)
+	stdout, stderr, err := runOVNretry(runner.nbctlPath, envVars, cmdArgs...)
 	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
 }
 
@@ -245,7 +347,7 @@ func RunOVNNbctl(args ...string) (string, string, error) {
 func RunOVNSbctlUnix(args ...string) (string, string, error) {
 	cmdArgs := []string{fmt.Sprintf("--timeout=%d", ovsCommandTimeout)}
 	cmdArgs = append(cmdArgs, args...)
-	stdout, stderr, err := runOVNretry(runner.sbctlPath, cmdArgs...)
+	stdout, stderr, err := runOVNretry(runner.sbctlPath, nil, cmdArgs...)
 	return strings.Trim(strings.TrimFunc(stdout.String(), unicode.IsSpace), "\""),
 		stderr.String(), err
 }
@@ -269,7 +371,7 @@ func RunOVNSbctlWithTimeout(timeout int, args ...string) (string, string,
 
 	cmdArgs = append(cmdArgs, fmt.Sprintf("--timeout=%d", timeout))
 	cmdArgs = append(cmdArgs, args...)
-	stdout, stderr, err := runOVNretry(runner.sbctlPath, cmdArgs...)
+	stdout, stderr, err := runOVNretry(runner.sbctlPath, nil, cmdArgs...)
 	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
 }
 
@@ -278,6 +380,48 @@ func RunOVNSbctl(args ...string) (string, string, error) {
 	return RunOVNSbctlWithTimeout(ovsCommandTimeout, args...)
 }
 
+// RunOVNCtl runs an ovn-ctl command.
+func RunOVNCtl(args ...string) (string, string, error) {
+	stdout, stderr, err := runOVNretry(runner.ovnctlPath, nil, args...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
+// RunOVNNBAppCtl runs an 'ovs-appctl -t nbdbCtlSockPath command'.
+func RunOVNNBAppCtl(args ...string) (string, string, error) {
+	var cmdArgs []string
+	cmdArgs = []string{
+		"-t",
+		runner.ovnRunDir + nbdbCtlSock,
+	}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := runOVNretry(runner.appctlPath, nil, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
+// RunOVNSBAppCtl runs an 'ovs-appctl -t sbdbCtlSockPath command'.
+func RunOVNSBAppCtl(args ...string) (string, string, error) {
+	var cmdArgs []string
+	cmdArgs = []string{
+		"-t",
+		runner.ovnRunDir + sbdbCtlSock,
+	}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := runOVNretry(runner.appctlPath, nil, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
+// RunOVNNorthAppCtl runs an 'ovs-appctl -t ovn-northd command'.
+func RunOVNNorthAppCtl(args ...string) (string, string, error) {
+	var cmdArgs []string
+	cmdArgs = []string{
+		"-t",
+		runner.ovnRunDir + northdCtlSock,
+	}
+	cmdArgs = append(cmdArgs, args...)
+	stdout, stderr, err := runOVNretry(runner.appctlPath, nil, cmdArgs...)
+	return strings.Trim(strings.TrimSpace(stdout.String()), "\""), stderr.String(), err
+}
+
 // RunIP runs a command via the iproute2 "ip" utility
 func RunIP(args ...string) (string, string, error) {
 	stdout, stderr, err := run(runner.ipPath, args...)
@@ -316,3 +460,21 @@ func RawExec(cmdPath string, args ...string) (string, string, error) {
 	stdout, stderr, err := run(cmdPath, args...)
 	return strings.TrimSpace(stdout.String()), stderr.String(), err
 }
+
+// AddNormalActionOFFlow replaces flows in the bridge with a NORMAL action flow
+func AddNormalActionOFFlow(bridgeName string) (string, string, error) {
+	args := []string{"-O", "OpenFlow13", "replace-flows", bridgeName, "-"}
+
+	stdin := &bytes.Buffer{}
+	stdin.Write([]byte("table=0,priority=0,actions=NORMAL\n"))
+
+	cmd := runner.exec.Command(runner.ofctlPath, args...)
+	cmd.SetStdin(stdin)
+	stdout, stderr, err := runCmd(cmd, runner.ofctlPath, args...)
+	return strings.Trim(stdout.String(), "\" \n"), stderr.String(), err
+}
+
+// GetOvnRunDir returns the OVN's rundir.
+func GetOvnRunDir() string {
+	return runner.ovnRunDir
+}
